# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Sequence,
    Union,
)

import numpy as np

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.validation import (
    ShapeT,
    bounded_by,
    shapeable,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    ArrayT,
    Checker,
    pipe,
    validated,
)

from .node_data import (
    Pwc,
    Tensor,
)

if TYPE_CHECKING:
    from boulderopal.graph._graph import Graph


def _check_ms_shapes(
    ion_count: int, lamb_dicke_parameters: np.ndarray, relative_detunings: np.ndarray
) -> None:
    """
    Check if the shapes of the Mølmer–Sørensen parameters are correct.

    The correct shapes for the input parameters of Mølmer–Sørensen gate are
    ``(3, ion_count, ion_count)`` for the Lamb–Dicke parameters and
    ``(3, ion_count)`` for the relative detunings.

    Parameters
    ----------
    ion_count : int
        The number of ions in the chain.
    lamb_dicke_parameters : np.ndarray
        The input values of the Lamb–Dicke parameters.
    relative_detunings : np.ndarray
        The input values of the relative detunings.
    """
    ld_shape = lamb_dicke_parameters.shape
    rd_shape = relative_detunings.shape

    Checker.VALUE(
        ld_shape == (3, ion_count, ion_count),
        "The Lamb–Dicke parameters must have shape (3, ion_count, ion_count).",
        {"ion_count": ion_count, "lamb_dicke_parameters.shape": ld_shape},
    )
    Checker.VALUE(
        rd_shape == (3, ion_count),
        "The relative detunings must have shape (3, ion_count).",
        {"ion_count": ion_count, "relative_detunings.shape": rd_shape},
    )


def _validate_drives(
    drives: Sequence[Optional[Pwc]], min_drive_count: int
) -> list[Pwc]:
    """
    Validate the drives for Mølmer–Sørensen operations.

    In particular, check that at least a minimum number of drives is provided,
    and that all drives have the same duration.

    Parameters
    ----------
    drives : list[Pwc or None]
        The drives.
    min_drive_count : int
        The minimum number of non-None drives.

    Returns
    -------
    list[Pwc]
        The non-None drives.
    """
    nonzero_drives = [drive for drive in drives if drive is not None]

    Checker.VALUE(
        len(nonzero_drives) >= min_drive_count,
        f"At least {min_drive_count} Pwc drive(s) must be provided.",
        {"drives": drives},
    )

    duration = sum(nonzero_drives[0].durations)
    Checker.VALUE(
        all(np.isclose(sum(drive.durations), duration) for drive in nonzero_drives),
        "All of the drives must have the same duration.",
    )
    return nonzero_drives


class IonsGraph:  # pylint: disable=unused-argument
    """
    Base class implementing Mølmer–Sørensen interaction graph methods.
    """

    def __init__(self, graph: "Graph") -> None:
        self._graph = graph

    @validated
    def ms_dephasing_robust_cost(
        self,
        drives: Sequence[
            Optional[Annotated[Pwc, pipe(shapeable, after=ShapeT.SIGNAL().no_batch())]]
        ],
        lamb_dicke_parameters: Annotated[np.ndarray, pipe(ArrayT.NUMERIC().ndim(3))],
        relative_detunings: Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(2))],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the cost for robust optimization of a Mølmer–Sørensen gate.

        Add the tensor that this function returns to the infidelity of your
        target operation to obtain a cost that you can use to create a
        Mølmer–Sørensen gate that is robust against dephasing noise. You can
        further multiply the robust cost by a scaling factor to weigh how much
        importance you give to the robustness compared to the original cost.

        Parameters
        ----------
        drives : list[Pwc or None]
            The piecewise-constant drives, :math:`\{\gamma_j\}`, one for each of the :math:`N` ions.
            Drive values must be in rad/s and durations must be in seconds.
            All drives must have the same total duration, but can have different segmentations.
            If an ion is not addressed, you can pass None for its drive (this leads to a
            more efficient calculation than passing a zero drive).
        lamb_dicke_parameters : np.ndarray
            The laser-ion coupling strength, :math:`\{\eta_{jkl}\}`.
            Its shape must be ``(3, N, N)``, where the dimensions indicate,
            respectively, axis, collective mode, and ion.
        relative_detunings : np.ndarray
            The difference :math:`\{\delta_{jk} = \nu_{jk} - \delta\}` (in Hz) between each motional
            mode frequency and the laser detuning from the qubit transition frequency
            :math:`\omega_0`. Its shape must be ``(3, N)``, where the dimensions indicate,
            respectively, axis and collective mode.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor(scalar, real)
            The cost term that you can use to optimize a Mølmer–Sørensen gate
            that is robust against dephasing noise. The cost is the sum of the
            square moduli of the time-averaged positions of the phase-space
            trajectories, weighted by the corresponding Lamb–Dicke parameters.

        See Also
        --------
        :func:`Graph.ions.ms_infidelity <ions.ms_infidelity>`
            Final operational infidelity of a Mølmer–Sørensen gate.
        :func:`boulderopal.ions.obtain_ion_chain_properties`
            Function to calculate the properties of an ion chain.

        Notes
        -----
        You can construct a Mølmer–Sørensen gate that is robust against
        dephasing noise by a combination of minimizing the time-averaged
        positions of the phase-space trajectories and imposing a symmetry in
        each ion's drive [1]_.

        The displacement of the :math:`l`-th ion in the :math:`k`-th mode of
        oscillation in dimension :math:`j` is [2]_

        .. math::
            \alpha_{jkl}(t) = \int_0^t d\tau \frac{\gamma_l(\tau)}{2}
                    \exp(i 2 \pi \delta_{jk} \tau) .

        For a gate of duration :math:`t_\text{gate}`, the time-averaged displacement is

        .. math::
            \langle \alpha_{jkl} \rangle = \frac{1}{t_\text{gate}}
                    \int_0^{t_\text{gate}} \alpha_{jkl}(t) \mathrm{d} t .

        This function returns the sum of the square moduli of the time-averaged
        positions multiplied by the corresponding Lamb–Dicke parameters. These
        parameters weight the time-averaged positions in the same way that the
        :math:`\alpha_{jkl}(t)` are weighted in the formula for the infidelity
        of a Mølmer–Sørensen gate (see :func:`Graph.ions.ms_infidelity <ions.ms_infidelity>`).

        In other words, the robust cost that this function returns is

        .. math::
            C_\text{robust} = \sum_{j=1}^{3} \sum_{k=1}^{N} \sum_{l=1}^{N}
                \left| \eta_{jkl} \langle \alpha_{jkl} \rangle \right|^2 .

        You can add this to the infidelity with the respect to the target gate
        to create the cost function that optimizes a gate that is also robust
        against dephasing. You can further multiply :math:`C_\text{robust}` by
        a scaling factor to weigh how much importance you give to robustness.

        References
        ----------
        .. [1] `A. R. Milne, C. L. Edmunds, C. Hempel, F. Roy, S. Mavadia,
                and M. J. Biercuk,
                Phys. Rev. Appl. 13, 024022 (2020).
                <https://doi.org/10.1103/PhysRevApplied.13.024022>`_
        .. [2] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
                M. R. Hush, and H. J. Slatyer,
                Adv. Quantum Technol. 3, 2000044 (2020).
                <https://doi.org/10.1002/qute.202000044>`_

        Examples
        --------
        Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust-molmer-
        sorensen-gates-for-trapped-ions>`_ user guide to find how to use this and related nodes.
        """

        _ = _validate_drives(drives, min_drive_count=1)
        _check_ms_shapes(
            ion_count=len(drives),
            lamb_dicke_parameters=lamb_dicke_parameters,
            relative_detunings=relative_detunings,
        )
        operation = create_operation(
            self.ms_dephasing_robust_cost, locals(), graph=self._graph
        )
        return Tensor(operation, shape=())

    @validated
    def ms_displacements(
        self,
        drives: Sequence[
            Optional[Annotated[Pwc, pipe(shapeable, after=ShapeT.SIGNAL().no_batch())]]
        ],
        lamb_dicke_parameters: Annotated[np.ndarray, pipe(ArrayT.NUMERIC().ndim(3))],
        relative_detunings: Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(2))],
        sample_times: Optional[
            Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).ge(0).ascend())]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the displacements for each mode and ion combination where ions are described by
        a Mølmer–Sørensen-type interaction.

        Use this function to calculate the displacements for each ion and each mode
        at the final time of the drives, or at the sample times that you provide.

        Parameters
        ----------
        drives : list[Pwc or None]
            The piecewise-constant drives, :math:`\{\gamma_j\}`, one for each of the :math:`N` ions.
            Drive values must be in rad/s and durations must be in seconds.
            All drives must have the same total duration, but can have different segmentations.
            If an ion is not addressed, you can pass None for its drive (this leads to a
            more efficient calculation than passing a zero drive).
        lamb_dicke_parameters : np.ndarray
            The laser-ion coupling strength, :math:`\{\eta_{jkl}\}`.
            Its shape must be ``(3, N, N)``, where the dimensions indicate,
            respectively, axis, collective mode, and ion.
        relative_detunings : np.ndarray
            The difference :math:`\{\delta_{jk} = \nu_{jk} - \delta\}` (in Hz) between each motional
            mode frequency and the laser detuning from the qubit transition frequency
            :math:`\omega_0`.  Its shape must be ``(3, N)``, where the dimensions indicate,
            respectively, axis and collective mode.
        sample_times : np.ndarray or None, optional
            The times (in seconds) at which to calculate the displacements, :math:`\{t_i\}`.
            If you provide it, it must be 1D, ordered, and contain at least one element.
            If you omit it, the displacements are only calculated at the final time of the drives.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor(complex)
            Displacements :math:`\{\eta_{jkl}\alpha_{jkl}(t_i)\}` for all mode-ion combinations.
            If you provide `sample_times`, the shape of the returned value is ``(T, 3, N, N_d)``,
            where the dimensions indicate, respectively, time, axis, collective mode, and ion.
            Otherwise, the shape is ``(3, N, N_d)``, with the outer time dimension removed.
            Here, ``N_d`` is the number of addressed ions (drives that are not None);
            the displacements of undriven ions are not returned (as they are zero).

        See Also
        --------
        :func:`Graph.ions.ms_infidelity <ions.ms_infidelity>`
            Final operational infidelity of a Mølmer–Sørensen gate.
        :func:`boulderopal.ions.obtain_ion_chain_properties`
            Function to calculate the properties of an ion chain.

        Notes
        -----
        This function calculates, at sample times :math:`\{t_i\}`, the contribution
        to the displacement of mode :math:`k` in dimension :math:`j` from ion :math:`l`,
        namely :math:`\eta_{jkl}\alpha_{jkl}(t_i)`, where

        .. math::
            \alpha_{jkl}(t) \equiv \int_0^t d\tau \frac{\gamma_l(\tau)}{2}
                \exp(i 2 \pi \delta_{jk} \tau) .

        You can calculate the state-dependent displacement by summing over the contributions
        from all ions. That is, using the displacement superoperator :math:`\mathcal{D}_{jk}`,
        the displacement in phase space for mode :math:`(j,k)` is

        .. math::
            \mathcal{D}_{jk} \left(\sum_{l=1}^N \sigma_{x,l}\eta_{jkl}\alpha_{jkl}(t_i) \right) .

        Examples
        --------
        Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust-molmer-sorensen
        -gates-for-trapped-ions>`_ user guide to find how to use this and related nodes.
        """
        nonzero_drives = _validate_drives(drives, min_drive_count=1)

        ion_count = len(drives)
        shape: tuple[int, ...] = (3, ion_count, len(nonzero_drives))
        _check_ms_shapes(
            ion_count=ion_count,
            lamb_dicke_parameters=lamb_dicke_parameters,
            relative_detunings=relative_detunings,
        )
        if sample_times is not None:
            duration = np.sum(nonzero_drives[0].durations)
            sample_times = bounded_by(
                sample_times, "sample_times", duration, "the drives duration"
            )
            shape = (len(sample_times),) + shape

        operation = create_operation(self.ms_displacements, locals(), graph=self._graph)
        return Tensor(operation, shape=shape)

    @validated
    def ms_infidelity(
        self,
        phases: Annotated[Union[Tensor, np.ndarray], pipe(shapeable)],
        displacements: Annotated[Union[Tensor, np.ndarray], pipe(shapeable)],
        target_phases: Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(2).square())],
        mean_phonon_numbers: Optional[
            Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(2))]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the final operational infidelity of the Mølmer–Sørensen gate.

        This function calculates the operational infidelity with respect to the target phases
        that you specify in the `target_phases` array. It can use the tensors returned from
        :func:`ms_phases` and :func:`ms_displacements` to calculate the infidelity tensor.

        Parameters
        ----------
        phases : np.ndarray(real) or Tensor(real)
            The acquired phases between ion pairs, :math:`\{\Phi_{ln}\}`.
            Its shape must be ``(N, N)`` without time samples or ``(T, N, N)`` with them,
            where ``T`` is the number of samples and ``N`` is the number of ions.
            For each sample, the `phases` array must be a strictly lower triangular matrix.
        displacements : np.ndarray(complex) or Tensor(complex)
            The motional displacements in phase-space, :math:`\{\eta_{jkl} \alpha_{jkl}\}`.
            Its shape must be ``(3, N, N_d)`` without time samples or ``(T, 3, N, N_d)`` with them,
            where the dimensions indicate, respectively, (time,) axis, collective mode, and ion.
            Here, ``N_d`` is the number of addressed ions.
        target_phases : np.ndarray(real)
            The target total relative phases between ion pairs, :math:`\{\Psi_{ln}\}`,
            as a strictly lower triangular matrix of shape ``(N, N)``.
        mean_phonon_numbers : np.ndarray or None, optional
            The mean phonon occupation of motional modes, :math:`\{\bar{n}_{jk}\}`.
            Its shape must be ``(3, N)``, where the dimensions indicate, respectively,
            axis and collective mode.
            If provided, must contain positive real numbers.
            If not provided, :math:`\bar{n}_{jk} = 0`, meaning no occupation of each mode.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor(real)
            A scalar or 1D tensor of infidelities with shape ``(T,)``.

        See Also
        --------
        :func:`Graph.ions.ms_dephasing_robust_cost <ions.ms_dephasing_robust_cost>`
            Cost for robust optimization of a Mølmer–Sørensen gate.
        :func:`Graph.ions.ms_displacements <ions.ms_displacements>`
            Displacements for each mode/ion combination.
        :func:`Graph.ions.ms_phases <ions.ms_phases>`
            Relative phases for all pairs of ions.

        Notes
        -----
        The infidelity is calculated according to [1]_

        .. math::
            1 - \mathcal{F}_\mathrm{av} = 1 - \left| \left( \prod_{n=1}^N \prod_{l=n+1}^N
                \cos ( \Phi_{ln} - \Psi_{ln}) \right)
                \left( 1 - \sum_{j=1}^3 \sum_{k=1}^N \sum_{l=1}^N \left[ |\eta_{jkl} \alpha_{jkl}|^2
                \left(\bar{n}_{jk}+\frac{1}{2} \right) \right] \right) \right|^2 ,

        which assumes that the displacements :math:`\alpha_{jkl}` are
        small and eliminates terms of the fourth or higher order in them.

        See the notes of :func:`Graph.ions.ms_phases <ions.ms_phases>` or
        :func:`Graph.ions.ms_phases_multitone <ions.ms_phases_multitone>` for the
        relevant definitions.

        References
        ----------
        .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
                M. R. Hush, and H. J. Slatyer,
                Adv. Quantum Technol. 3, 2000044 (2020).
                <https://doi.org/10.1002/qute.202000044>`_

        Examples
        --------
        Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust-molmer-
        sorensen-gates-for-trapped-ions>`_ user guide to find how to use this and related nodes.
        """

        ion_count = target_phases.shape[-1]
        Checker.VALUE(
            len(phases.shape) in (2, 3) and phases.shape[-2:] == (ion_count, ion_count),
            "The shape of phases must be (ion_count, ion_count) or"
            " (sample_count, ion_count, ion_count).",
            {"ion_count": ion_count, "phases.shape": phases.shape},
        )
        Checker.VALUE(
            len(displacements.shape) in (3, 4)
            and displacements.shape[-3:-1] == (3, ion_count),
            "The shape of displacements must be (3, ion_count, driven_ion_count) or"
            " (sample_count, 3, ion_count, driven_ion_count).",
            {"ion_count": ion_count, "displacements.shape": displacements.shape},
        )
        Checker.VALUE(
            phases.shape[:-2] == displacements.shape[:-3],
            "The number of samples in phases and displacements must be the same.",
            {"phases.shape": phases.shape, "displacements.shape": displacements.shape},
        )
        Checker.VALUE(
            target_phases.shape == (ion_count, ion_count),
            "The shape of target_phases must be (ion_count, ion_count).",
            {"ion_count": ion_count, "target_phases.shape": target_phases.shape},
        )
        Checker.VALUE(
            np.allclose(target_phases, np.tril(target_phases, k=-1)),
            "The target_phases matrix must be strictly lower-triangular.",
        )
        if mean_phonon_numbers is not None:
            Checker.VALUE(
                mean_phonon_numbers.shape == (3, ion_count),
                "The shape of mean_phonon_numbers must be (3, ion_count).",
                {
                    "ion_count": ion_count,
                    "mean_phonon_number.shape": mean_phonon_numbers.shape,
                },
            )

        shape = phases.shape[:-2]
        operation = create_operation(self.ms_infidelity, locals(), graph=self._graph)
        return Tensor(operation, shape=shape)

    @validated
    def ms_phases(
        self,
        drives: Sequence[
            Optional[Annotated[Pwc, pipe(shapeable, after=ShapeT.SIGNAL().no_batch())]]
        ],
        lamb_dicke_parameters: Annotated[np.ndarray, pipe(ArrayT.NUMERIC().ndim(3))],
        relative_detunings: Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(2))],
        sample_times: Optional[
            Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).ge(0).ascend())]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the relative phases for all pairs of ions described by a Mølmer–Sørensen-type
        interaction when single-tone individually-addressed laser beams are used.

        Use this function to calculate the acquired phases for all ion pairs
        at the final time of the drives, or at the sample times that you provide.

        Parameters
        ----------
        drives : list[Pwc or None]
            The piecewise-constant drives, :math:`\{\gamma_j\}`, one for each of the :math:`N` ions.
            Drive values must be in rad/s and durations must be in seconds.
            All drives must have the same total duration, but can have different segmentations.
            This list must contain at least two elements (your system must have at least two ions).
            If an ion is not addressed, you can pass None for its drive (this leads to a
            more efficient calculation than passing a zero drive).
        lamb_dicke_parameters : np.ndarray
            The laser-ion coupling strength, :math:`\{\eta_{jkl}\}`.
            Its shape must be ``(3, N, N)``, where the dimensions indicate,
            respectively, axis, collective mode, and ion.
        relative_detunings : np.ndarray
            The difference :math:`\{\delta_{jk} = \nu_{jk} - \delta\}` (in Hz) between each motional
            mode frequency and the laser detuning from the qubit transition frequency
            :math:`\omega_0`.  Its shape must be ``(3, N)``, where the dimensions indicate,
            respectively, axis and collective mode.
        sample_times : np.ndarray or None, optional
            The times (in seconds) at which to calculate the relative phases, :math:`\{t_i\}`.
            If you provide it, it must be 1D, ordered, and contain at least one element.
            If you omit it, the phases are only calculated at the final time of the drives.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor(real)
            Acquired phases :math:`\{\phi_{jk}(t_i) + \phi_{kj}(t_i)\}` for all ion pairs.
            If you provide `sample_times`, the shape of the returned value is ``(T, N, N)``,
            where the first dimension indicates time, and the second and third dimensions indicate
            ions. Otherwise, the shape is ``(N, N)``, with the outer time dimension removed.
            The relative phases are stored as a strictly lower triangular matrix.
            See the notes part for details.

        See Also
        --------
        :func:`Graph.ions.ms_infidelity <ions.ms_infidelity>`
            Final operational infidelity of a Mølmer–Sørensen gate.
        :func:`Graph.ions.ms_phases_multitone <ions.ms_phases_multitone>`
            Corresponding operation for a global multitone beam.
        :func:`boulderopal.ions.obtain_ion_chain_properties`
            Function to calculate the properties of an ion chain.

        Notes
        -----
        The internal and motional Hamiltonian of :math:`N` ions is

        .. math::
            H_0 = \sum_{j=1}^{3} \sum_{k=1}^{N} \hbar\nu_{jk} \left(a_{jk}^\dagger a_{jk}
                + \frac{1}{2}\right) + \sum_{l=1}^N \frac{\hbar \omega_0}{2} \sigma_{z,l} ,

        where :math:`j` indicates axis dimension (:math:`x`, :math:`y`, or :math:`z`),
        :math:`k` indicates collective mode, :math:`a_{jk}` is the annihilation operator,
        and :math:`\sigma_{z,l}` is the Pauli :math:`Z` operator for ion :math:`l`.

        The interaction Hamiltonian for Mølmer–Sørensen-type
        operations in the rotating frame with respect to :math:`H_0` is

        .. math::
            H_I(t) = i\hbar \sum_{j=1}^{3} \sum_{k=1}^{N} \sum_{l=1}^N
                \sigma_{x,l} \left(-\beta_{jkl}^*(t)a_{jk} + \beta_{jkl}(t) a_{jk}^\dagger\right) ,

        where :math:`\sigma_{x,l}` is the Pauli :math:`X` operator for ion :math:`l` and

        .. math::
            \beta_{jkl}(t) = \eta_{jkl} \frac{\gamma_l(t)}{2} \exp(i 2 \pi \delta_{jk} t)

        indicates the coupling between ion :math:`l` and motional mode :math:`(j,k)`.

        The corresponding unitary operation is given by [1]_

        .. math::
            U(t) = \exp\left[ \sum_{l=1}^N \sigma_{x,l} B_l(t)
                + i\sum_{l=1}^N\sum_{n=1}^{l-1} (\phi_{ln}(t) + \phi_{nl}(t))
                \sigma_{x,l} \sigma_{x,n} \right] ,

        where

        .. math::
            B_l(t) &\equiv \sum_{j=1}^{3} \sum_{k=1}^{N}
                \left(\eta_{jkl}\alpha_{jkl}(t)a_{jk}^\dagger
                - \eta_{jkl}^{\ast}\alpha_{jkl}^\ast(t)a_{jk} \right) ,

            \phi_{ln}(t) &\equiv \mathrm{Im} \left[ \sum_{j=1}^{3} \sum_{k=1}^{N}
                \int_{0}^{t} d \tau_1 \int_{0}^{\tau_1} d \tau_2
                \beta_{jkl}(\tau_1)\beta_{jkn}^{\ast}(\tau_2) \right] ,

            \alpha_{jkl}(t) &\equiv \int_0^t d\tau \frac{\gamma_l(\tau)}{2}
                \exp(i 2 \pi \delta_{jk} \tau) .

        This function calculates the relative phases for all ions pairs
        at sample times :math:`\{t_i\}`,

        .. math::
            \Phi_{ln}(t_i) = \phi_{ln}(t_i) + \phi_{nl}(t_i),

        and stores them in a strictly lower triangular matrix.
        That is, :math:`\Phi_{ln}(t_i)` with :math:`l > n`
        gives the relative phase between ions :math:`l` and :math:`n`,
        while :math:`\Phi_{ln}(t_i) = 0` for :math:`l \leq n`.

        References
        ----------
        .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
                M. R. Hush, and H. J. Slatyer,
                Adv. Quantum Technol. 3, 2000044 (2020).
                <https://doi.org/10.1002/qute.202000044>`_

        Examples
        --------
        Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust-molmer-
        sorensen-gates-for-trapped-ions>`_ user guide to find how to use this and related nodes.
        """

        nonzero_drives = _validate_drives(drives, min_drive_count=2)

        ion_count = len(drives)
        shape: tuple[int, ...] = (ion_count, ion_count)
        _check_ms_shapes(
            ion_count=ion_count,
            lamb_dicke_parameters=lamb_dicke_parameters,
            relative_detunings=relative_detunings,
        )
        if sample_times is not None:
            duration = np.sum(nonzero_drives[0].durations)
            sample_times = bounded_by(
                sample_times, "sample_times", duration, "the drives duration"
            )
            shape = (len(sample_times),) + shape

        operation = create_operation(self.ms_phases, locals(), graph=self._graph)
        return Tensor(operation, shape=shape)

    @validated
    def ms_phases_multitone(
        self,
        drives: List[Annotated[Pwc, pipe(after=ShapeT.SIGNAL().no_batch())]],
        lamb_dicke_parameters: Annotated[np.ndarray, pipe(ArrayT.NUMERIC().ndim(4))],
        relative_detunings: Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(3))],
        sample_times: Optional[
            Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).ge(0).ascend())]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the relative phases for all pairs of ions described by a
        Mølmer–Sørensen-type interaction where the ions are being addressed by
        a multitone global beam.

        Use this function to calculate the acquired phases for all ion pairs
        at the final time of the drives, or at the sample times that you provide.

        Parameters
        ----------
        drives : list[Pwc]
            The piecewise-constant drives, :math:`\{\gamma_j\}`,
            one for each of the :math:`M` tones of the global beam.
            Drive values must be in rad/s and durations must be in seconds.
            All drives must have the same total duration, but can have different segmentations.
        lamb_dicke_parameters : np.ndarray
            The laser-ion coupling strength, :math:`\{\eta_{\xi jkl}\}`.
            Its shape must be ``(M, 3, N, N)``, where the dimensions indicate,
            respectively, tone of the global beam, axis, collective mode, and ion.
        relative_detunings : np.ndarray
            The difference :math:`\{\delta_{\xi jk} = \nu_{jk} - \delta_\xi \}` (in Hz) between
            each motional mode frequency and the laser detunings for each tone from the qubit
            transition frequency :math:`\omega_0`. Its shape must be ``(M, 3, N)``, where the
            dimensions indicate, respectively, tone of the global beam, axis, and collective mode.
        sample_times : np.ndarray or None, optional
            The times (in seconds) at which to calculate the relative phases, :math:`\{t_i\}`.
            If you provide it, it must be 1D, ordered, and contain at least one element.
            If you omit it, the phases are only calculated at the final time of the drives.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor(real)
            Acquired phases :math:`\{\phi_{jk}(t_i) + \phi_{kj}(t_i)\}` for all ion pairs.
            If you provide `sample_times`, the shape of the returned value is ``(T, N, N)``,
            where the first dimension indicates time, and the second and third dimensions indicate
            ions. Otherwise, the shape is ``(N, N)``, with the outer time dimension removed.
            The relative phases are stored as a strictly lower triangular matrix.
            See the notes part for details.

        See Also
        --------
        :func:`Graph.ions.ms_infidelity <ions.ms_infidelity>`
            Final operational infidelity of a Mølmer–Sørensen gate.
        :func:`Graph.ions.ms_phases <ions.ms_phases>`
            Corresponding operation for single-tone individually-addressed beams.
        :func:`boulderopal.ions.obtain_ion_chain_properties`
            Function to calculate the properties of an ion chain.

        Notes
        -----
        The interaction Hamiltonian for Mølmer–Sørensen-type
        operations in the rotating frame for a multitone global beam is

        .. math::
            H_I(t) = i\hbar \sum_{\xi=1}^M \sum_{j=1}^{3} \sum_{k=1}^{N} \sum_{l=1}^N \sigma_{x,l}
                \left(-\beta_{\xi jkl}^*(t)a_{jk} + \beta_{\xi jkl}(t) a_{jk}^\dagger\right) ,

        where :math:`\sigma_{x,l}` is the Pauli :math:`X` operator for the ion :math:`l` and

        .. math::

            \beta_{\xi jkl}(t) = \eta_{\xi jkl} \frac{\gamma_\xi(t)}{2}
                \exp(i 2 \pi \delta_{\xi jk} t)

        indicates the coupling between ion :math:`l` and motional mode :math:`(\xi,j,k)`.

        The corresponding unitary operation is given by

        .. math::
            U(t) = \exp\left[ \sum_{l=1}^N \sigma_{x,l} B_l(t)
                    + i\sum_{l=1}^N \sum_{n=1}^{l-1} (\phi_{ln}(t) + \phi_{nl}(t))
                    \sigma_{x,l} \sigma_{x,n} \right] ,

        where

        .. math::
            B_l(t) &\equiv \sum_{\xi=1}^M \sum_{j=1}^{3} \sum_{k=1}^{N}
                \left(\eta_{\xi jkl}\alpha_{\xi jkl}(t)a_{jk}^\dagger
                - \eta_{\xi jkl}^{\ast}\alpha_{\xi jkl}^\ast(t)a_{jk} \right) ,

            \phi_{ln}(t) &\equiv \mathrm{Im} \left[
                \sum_{\xi=1}^M \sum_{\chi=1}^M \sum_{j=1}^{3} \sum_{k=1}^{N}
                \int_{0}^{t} d \tau_1 \int_{0}^{\tau_1} d \tau_2
                \beta_{\xi jkl}(\tau_1)\beta_{\chi jkn}^{\ast}(\tau_2) \right] ,

            \alpha_{\xi jkl}(t) &\equiv \int_0^t d\tau \frac{\gamma_\xi(\tau)}{2}
                \exp(i 2 \pi \delta_{\xi jk} \tau).

        This function calculates the relative phases for all ions pairs
        at sample times :math:`\{t_i\}`,

        .. math::
            \Phi_{ln}(t_i) = \phi_{ln}(t_i) + \phi_{nl}(t_i),

        and stores them in a strictly lower triangular matrix.
        That is, :math:`\Phi_{ln}(t_i)` with :math:`l > n`
        gives the relative phase between ions :math:`l` and :math:`n`,
        while :math:`\Phi_{ln}(t_i) = 0` for :math:`l \leq n`.
        """

        _ = _validate_drives(drives, min_drive_count=1)

        ld_shape = lamb_dicke_parameters.shape
        rd_shape = relative_detunings.shape

        tone_count = len(drives)
        ion_count = ld_shape[-1]
        shape: tuple[int, ...] = (ion_count, ion_count)

        Checker.VALUE(
            ld_shape == (tone_count, 3, ion_count, ion_count),
            "The Lamb–Dicke parameters must have shape (tone_count, 3, ion_count, ion_count).",
            {
                "ion_count": ion_count,
                "tone_count": tone_count,
                "lamb_dicke_parameters.shape": ld_shape,
            },
        )
        Checker.VALUE(
            rd_shape == (tone_count, 3, ion_count),
            "The relative detunings must have shape (tone_count, 3, ion_count).",
            {
                "ion_count": ion_count,
                "tone_count": tone_count,
                "relative_detunings.shape": rd_shape,
            },
        )

        if sample_times is not None:
            duration = np.sum(drives[0].durations)
            sample_times = bounded_by(
                sample_times, "sample_times", duration, "the drives duration"
            )
            shape = (len(sample_times),) + shape

        operation = create_operation(
            self.ms_phases_multitone, locals(), graph=self._graph
        )
        return Tensor(operation, shape=shape)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    List,
    Optional,
    Union,
)

import numpy as np
from pydantic import Field

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.node_data import (
    ConvolutionKernel,
    Pwc,
    Stf,
    Tensor,
)
from boulderopal._nodes.validation import (
    ArrayT,
    ShapeT,
    non_negative_scalar,
    positive_scalar,
    scalar_or_shapeable,
    shapeable,
    starts_with_zero,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    Checker,
    ScalarT,
    pipe,
    validated,
)


class StfGraph:  # pylint: disable=unused-argument
    """
    Base class implementing Stf graph methods.
    """

    @validated
    def stf_operator(
        self,
        signal: Annotated[Stf, pipe(after=ShapeT.SIGNAL())],
        operator: Annotated[
            Union[np.ndarray, Tensor],
            pipe(shapeable, after=ShapeT.OPERATOR().no_batch()),
        ],
    ) -> Stf:
        r"""
        Create a constant operator multiplied by a sampleable signal.

        Parameters
        ----------
        signal : Stf
            A sampleable function representing the signal :math:`a(t)`
            or a batch of sampleable functions.
        operator : np.ndarray or Tensor
            The operator :math:`A`. It must have two equal dimensions.

        Returns
        -------
        Stf
            The sampleable operator :math:`a(t)A` (or batch of sampleable operators, if
            you provide a batch of signals).

        See Also
        --------
        Graph.constant_stf_operator : Create a constant `Stf` operator.
        Graph.pwc_operator : Corresponding operation for `Pwc`\s.
        Graph.stf_sum : Sum multiple `Stf`\s.

        Notes
        -----
        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.
        """

        operation = create_operation(self.stf_operator, locals())
        return Stf(
            operation, value_shape=operator.shape, batch_shape=signal.batch_shape
        )

    @validated
    def constant_stf_operator(
        self,
        operator: Annotated[
            Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.OPERATOR())
        ],
    ) -> Stf:
        r"""
        Create a constant operator.

        Parameters
        ----------
        operator : np.ndarray or Tensor
            The operator :math:`A`, or a batch of operators. It must have at
            least two dimensions, and its last two dimensions must be equal.

        Returns
        -------
        Stf
            The operator :math:`t\mapsto A` (or batch of
            operators, if you provide a batch of operators).

        See Also
        --------
        Graph.constant_pwc_operator : Corresponding operation for `Pwc`\s.
        Graph.constant_stf : Create a batch of constant `Stf`\s.
        Graph.stf_operator : Create an `Stf` operator.

        Notes
        -----
        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.
        """
        operation = create_operation(self.constant_stf_operator, locals())
        return Stf(
            operation, value_shape=operator.shape[-2:], batch_shape=operator.shape[:-2]
        )

    @validated
    def constant_stf(
        self,
        constant: Annotated[
            Union[float, complex, np.ndarray, Tensor], pipe(scalar_or_shapeable)
        ],
        batch_dimension_count: Annotated[int, pipe(ScalarT.INT().ge(0))] = 0,
    ) -> Stf:
        r"""
        Create a constant sampleable tensor-valued function of time.

        Parameters
        ----------
        constant : number or np.ndarray or Tensor
            The constant value :math:`c` of the function.
            To create a batch of :math:`B_1 \times \ldots \times B_n` constant
            functions of shape :math:`D_1 \times \ldots \times D_m`, provide this `constant`
            parameter as an object of shape
            :math:`B_1\times\ldots\times B_n\times D_1\times\ldots\times D_m`.
        batch_dimension_count : int, optional
            The number of batch dimensions, :math:`n`, in `constant`.
            If provided, the first :math:`n` dimensions of `constant` are considered
            batch dimensions. Defaults to 0, which corresponds to no batch.

        Returns
        -------
        Stf
           An Stf representing the constant function :math:`f(t) = c` for all time
           :math:`t` (or a batch of functions, if you provide `batch_dimension_count`).

        See Also
        --------
        Graph.constant_pwc : Corresponding operation for `Pwc`\s.
        Graph.constant_stf_operator : Create a constant sampleable function from operators.
        Graph.identity_stf : Create an `Stf` representing the identity function.

        Notes
        -----
        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.
        """

        shape = getattr(constant, "shape", ())
        Checker.VALUE(
            len(shape) >= batch_dimension_count,
            "The number of batch dimensions must not be larger than the number "
            "of dimensions of the input constant.",
            {"constant shape": shape, "batch_dimension_count": batch_dimension_count},
        )
        return Stf(
            operation=create_operation(self.constant_stf, locals()),
            value_shape=shape[batch_dimension_count:],
            batch_shape=shape[:batch_dimension_count],
        )

    @validated
    def stf_sum(self, terms: Annotated[List[Stf], Field(min_length=1)]) -> Stf:
        r"""
        Create the sum of multiple sampleable functions.

        Parameters
        ----------
        terms : list[Stf]
            The individual sampleable function :math:`\{v_j(t)\}` to sum.

        Returns
        -------
        Stf
            The sampleable function of time :math:`\sum_j v_j(t)`. It has the same
            shape as each of the `terms` that you provide.

        See Also
        --------
        Graph.hermitian_part : Hermitian part of an `Stf` operator.
        Graph.pwc_sum : Corresponding operation for `Pwc`\s.
        Graph.stf_operator : Create an `Stf` operator.

        Notes
        -----
        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.
        """

        for attr in ["value_shape", "batch_shape"]:
            Checker.VALUE(
                len(set(getattr(term, attr) for term in terms)) == 1,
                f"All the terms must have the same {attr}.",
            )
        operation = create_operation(self.stf_sum, locals())
        return Stf(
            operation,
            value_shape=terms[0].value_shape,
            batch_shape=terms[0].batch_shape,
        )

    @validated
    def discretize_stf(
        self,
        stf: Stf,
        duration: Annotated[float, pipe(ScalarT.REAL().gt(0))],
        segment_count: Annotated[int, pipe(ScalarT.INT().gt(0))],
        sample_count_per_segment: Annotated[int, pipe(ScalarT.INT().gt(0))] = 1,
        *,
        name: Optional[str] = None,
    ) -> Pwc:
        r"""
        Create a piecewise-constant function by discretizing a sampleable function.

        Use this function to create a piecewise-constant approximation to a sampleable
        function (obtained, for example, by filtering an initial
        piecewise-constant function).

        Parameters
        ----------
        stf : Stf
            The sampleable function :math:`v(t)` to discretize. The values of the
            function can have any shape. You can also provide a batch of
            functions, in which case the discretization is applied to each
            element of the batch.
        duration : float
            The duration :math:`\tau` over which discretization should be
            performed. The resulting piecewise-constant function has this
            duration.
        segment_count : int
            The number of segments :math:`N` in the resulting piecewise-constant
            function.
        sample_count_per_segment : int, optional
            The number of samples :math:`M` of the sampleable function to take when
            calculating the value of each segment in the discretization. Defaults
            to 1.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The piecewise-constant function :math:`w(t)` obtained by discretizing
            the sampleable function (or batch of piecewise-constant functions, if
            you provided a batch of sampleable functions).

        See Also
        --------
        Graph.convolve_pwc : Create an `Stf` by convolving a `Pwc` with a kernel.
        Graph.filter_and_resample_pwc : Filter a `Pwc` with a sinc filter and resample it.
        Graph.identity_stf : Create an `Stf` representing the identity function.
        Graph.sample_stf : Sample an `Stf` at given times.

        Notes
        -----
        The resulting function :math:`w(t)` is piecewise-constant with :math:`N`
        segments, meaning it has segment values :math:`\{w_n\}` such that
        :math:`w(t)=w_n` for :math:`t_{n-1}\leq t\leq t_n`, where :math:`t_n= n \tau/N`.

        Each segment value :math:`w_n` is the average of samples of :math:`v(t)`
        at the midpoints of :math:`M` equally sized subsegments between
        :math:`t_{n-1}` and :math:`t_n`:

        .. math::
            w_n = \frac{1}{M}
            \sum_{m=1}^M v\left(t_{n-1} + \left(m-\tfrac{1}{2}\right) \frac{\tau}{MN} \right).

        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.

        Examples
        --------
        Create discretized Gaussian signal.

        >>> times = graph.identity_stf()
        >>> gaussian_signal = graph.exp(- (times - 5e-6) ** 2 / 2e-6 ** 2) / 2e-6
        >>> discretized_gamma_signal = graph.discretize_stf(
        ...     stf=gaussian_signal, duration=10e-6, segment_count=256, name="gamma"
        ... )
        >>> discretized_gamma_signal
        <Pwc: name="gamma", operation_name="discretize_stf", value_shape=(), batch_shape=()>

        Refer to the `How to create dephasing and amplitude robust single-qubit gates
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-create-dephasing-and-amplitude-
        robust-single-qubit-gates>`_ user guide to find the example in context.
        """

        durations = duration / segment_count * np.ones(segment_count)
        return Pwc(
            operation=create_operation(self.discretize_stf, locals()),
            durations=durations,
            value_shape=stf.value_shape,
            batch_shape=stf.batch_shape,
        )

    @validated
    def time_evolution_operators_stf(
        self,
        hamiltonian: Annotated[Stf, pipe(after=ShapeT.OPERATOR())],
        sample_times: Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).ge(0).ascend())],
        evolution_times: Optional[
            Annotated[
                np.ndarray,
                pipe(ArrayT.REAL().ndim(1).ge(0).ascend(), after=starts_with_zero),
            ]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Calculate the time-evolution operators for a system defined by an STF Hamiltonian by using a
        4th order Runge–Kutta method.

        Parameters
        ----------
        hamiltonian : Stf
            The control Hamiltonian, or batch of control Hamiltonians.
        sample_times : np.ndarray
            The N times at which you want to sample the unitaries. Must be ordered and contain
            at least one element. If you don't provide `evolution_times`, `sample_times` must
            start with 0.
        evolution_times : np.ndarray or None, optional
            The times at which the Hamiltonian should be sampled for the Runge–Kutta integration.
            If you provide it, must start with 0 and be ordered.
            If you don't provide it, the `sample_times` are used for the integration.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            Tensor of shape ``(..., N, D, D)``, representing the unitary time evolution.
            The n-th element (along the -3 dimension) represents the unitary (or batch of unitaries)
            from t = 0 to ``sample_times[n]``.

        See Also
        --------
        Graph.time_evolution_operators_pwc : Corresponding operation for `Pwc` Hamiltonians.

        Notes
        -----
        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.

        Examples
        --------
        Simulate the dynamics of a qubit, where a simple Gaussian drive rotate the qubit
        along the x-axis.

        >>> duration = np.pi
        >>> initial_state = np.array([1, 0])
        >>> sigma_x = np.array([[0, 1], [1, 0]])
        >>> time = graph.identity_stf()
        >>> gaussian_drive = graph.exp(-(time ** 2))
        >>> hamiltonian = gaussian_drive * sigma_x * np.sqrt(np.pi) / 2
        >>> graph.time_evolution_operators_stf(
        ...     hamiltonian=hamiltonian,
        ...     sample_times=[duration],
        ...     evolution_times=np.linspace(0, duration, 200),
        ...     name="unitaries",
        ... )
        <Tensor: name="unitaries", operation_name="time_evolution_operators_stf", shape=(1, 2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="unitaries")
        >>> result["output"]["unitaries"]["value"].dot(initial_state)
        array([[0.70711169 + 0.0j, 0.0 - 0.70710187j]])
        """
        if evolution_times is None:
            Checker.VALUE(
                sample_times[0] == 0,
                "If you don't provide evolution times, the first of the sample times must be zero.",
            )
        shape = hamiltonian.batch_shape + (len(sample_times),) + hamiltonian.value_shape
        operation = create_operation(self.time_evolution_operators_stf, locals())
        return Tensor(operation, shape=shape)

    @validated
    def convolve_pwc(self, pwc: Pwc, kernel: ConvolutionKernel) -> Stf:
        r"""
        Create the convolution of a piecewise-constant function with a kernel.

        Parameters
        ----------
        pwc : Pwc
            The piecewise-constant function :math:`\alpha(t)` to convolve. You
            can provide a batch of functions, in which case the convolution is
            applied to each element of the batch.
        kernel : ConvolutionKernel
            The node representing the kernel :math:`K(t)`.

        Returns
        -------
        Stf
            The sampleable function representing the signal :math:`(\alpha * K)(t)`
            (or batch of signals, if you provide a batch of functions).

        See Also
        --------
        Graph.discretize_stf : Discretize an `Stf` into a `Pwc`.
        Graph.filter_and_resample_pwc : Filter a `Pwc` with a sinc filter and resample it.
        Graph.gaussian_convolution_kernel :
            Create a convolution kernel representing a normalized Gaussian.
        Graph.pwc : Create piecewise-constant functions.
        Graph.sample_stf : Sample an `Stf` at given times.
        Graph.sinc_convolution_kernel : Create a convolution kernel representing the sinc function.

        Notes
        -----
        The convolution is

        .. math::
            (\alpha * K)(t) \equiv
            \int_{-\infty}^\infty \alpha(\tau) K(t-\tau) d\tau.

        Convolution in the time domain is equivalent to multiplication in the
        frequency domain, so this function can be viewed as applying a linear
        time-invariant filter (specified via its time domain kernel :math:`K(t)`)
        to :math:`\alpha(t)`.

        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.

        Examples
        --------
        Filter a piecewise-constant signal using a Gaussian convolution kernel.

        >>> gaussian_kernel = graph.gaussian_convolution_kernel(std=1.0, offset=3.0)
        >>> gaussian_kernel
        <ConvolutionKernel: operation_name="gaussian_convolution_kernel">
        >>> pwc_signal
        <Pwc: name="alpha", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        >>> filtered_signal = graph.convolve_pwc(pwc=pwc_signal, kernel=gaussian_kernel)
        >>> filtered_signal
        <Stf: operation_name="convolve_pwc", value_shape=(), batch_shape=()>

        Refer to the `How to add smoothing and band-limits to optimized controls
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-add-smoothing-and-band-limits-to-optimized-controls>`_
        user guide to find the example in context.
        """
        operation = create_operation(self.convolve_pwc, locals())
        return Stf(operation, value_shape=pwc.value_shape, batch_shape=pwc.batch_shape)

    @validated
    def sinc_convolution_kernel(
        self, cutoff_frequency: Annotated[Union[float, Tensor], pipe(positive_scalar)]
    ) -> ConvolutionKernel:
        r"""
        Create a convolution kernel representing the sinc function.

        Use this kernel to eliminate angular frequencies above a certain cutoff.

        Parameters
        ----------
        cutoff_frequency : float or Tensor
            Upper limit :math:`\omega_c` of the range of angular frequencies that you want
            to preserve. The filter eliminates components of the signal that have
            higher angular frequencies.

        Returns
        -------
        ConvolutionKernel
            A node representing the sinc function to use in a convolution.

        See Also
        --------
        Graph.convolve_pwc : Create an `Stf` by convolving a `Pwc` with a kernel.
        Graph.filter_and_resample_pwc : Filter a `Pwc` with a sinc filter and resample it.
        Graph.gaussian_convolution_kernel :
            Create a convolution kernel representing a normalized Gaussian.

        Notes
        -----
        The sinc kernel that this node represents is defined as

        .. math::
            K(t) = \frac{\sin(\omega_c t)}{\pi t}.

        In the frequency domain, the sinc function is constant in the range
        :math:`[-\omega_c, \omega_c]` and zero elsewhere. The filter it represents therefore
        passes angular frequencies only in that range.

        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.

        Examples
        --------
        Filter a signal by convolving it with a sinc kernel.

        >>> sinc_kernel = graph.sinc_convolution_kernel(cutoff_frequency=300e6)
        >>> sinc_kernel
        <ConvolutionKernel: operation_name="sinc_convolution_kernel">
        >>> pwc_signal
        <Pwc: name="pwc_signal_#1", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        >>> filtered_signal = graph.convolve_pwc(pwc=pwc_signal, kernel=sinc_kernel)
        >>> filtered_signal
        <Stf: operation_name="convolve_pwc", value_shape=(), batch_shape=()>

        Refer to the `How to create leakage-robust single-qubit gates
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-create-leakage-robust-single-qubit-gates>`_
        user guide to find the example in context.
        """
        return ConvolutionKernel(
            create_operation(self.sinc_convolution_kernel, locals())
        )

    @validated
    def gaussian_convolution_kernel(
        self,
        std: Annotated[Union[float, Tensor], pipe(positive_scalar)],
        offset: Annotated[Union[float, Tensor], pipe(non_negative_scalar)] = 0,
    ) -> ConvolutionKernel:
        r"""
        Create a convolution kernel representing a normalized Gaussian.

        Use this kernel to allow angular frequencies in the range roughly determined by
        its width, and progressively suppress components outside that range.

        Parameters
        ----------
        std : float or Tensor
            Standard deviation :math:`\sigma` of the Gaussian in the time domain.
            The standard deviation in the frequency domain is its inverse, so that
            a high value of this parameter lets fewer angular frequencies pass.
        offset : float or Tensor, optional
            Center :math:`\mu` of the Gaussian distribution in the time domain.
            Use this to offset the signal in time. Defaults to 0.

        Returns
        -------
        ConvolutionKernel
            A node representing a Gaussian function to use in a convolution.

        See Also
        --------
        Graph.convolve_pwc : Create an `Stf` by convolving a `Pwc` with a kernel.
        Graph.sinc_convolution_kernel : Create a convolution kernel representing the sinc function.

        Notes
        -----
        The Gaussian kernel that this node represents is defined as:

        .. math::
            K(t) = \frac{e^{-(t-\mu)^2/(2\sigma^2)}}{\sqrt{2\pi\sigma^2}}.

        In the frequency domain, this Gaussian has standard deviation
        :math:`\omega_c= \sigma^{-1}`. The filter it represents therefore
        passes angular frequencies roughly in the range :math:`[-\omega_c, \omega_c]`.

        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.

        Examples
        --------
        Filter a signal by convolving it with a Gaussian kernel.

        >>> gaussian_kernel = graph.gaussian_convolution_kernel(std=1.0, offset=3.0)
        >>> gaussian_kernel
        <ConvolutionKernel: operation_name="gaussian_convolution_kernel">
        >>> signal
        <Pwc: name="alpha", operation_name="pwc_signal", value_shape=(), batch_shape=()>
        >>> filtered_signal = graph.convolve_pwc(pwc=signal, kernel=gaussian_kernel)
        >>> filtered_signal
        <Stf: operation_name="convolve_pwc", value_shape=(), batch_shape=()>

        Refer to the `How to characterize a transmission line using a qubit as a probe
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-characterize-a-transmission-
        line-using-a-qubit-as-a-probe>`_ user guide to find the example in context.
        """
        return ConvolutionKernel(
            create_operation(self.gaussian_convolution_kernel, locals())
        )

    @validated
    def sample_stf(
        self,
        stf: Stf,
        sample_times: Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).ge(0).ascend())],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Sample an Stf at the given times.

        Parameters
        ----------
        stf : Stf
            The Stf to sample.
        sample_times : np.ndarray
            The times at which you want to sample the Stf. Must be ordered and contain
            at least one element.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The values of the Stf at the given times.

        See Also
        --------
        Graph.constant_stf_operator : Create a constant `Stf` operator.
        Graph.discretize_stf : Discretize an `Stf` into a `Pwc`.
        Graph.identity_stf : Create an `Stf` representing the identity function.
        Graph.sample_pwc : Sample a `Pwc` at given times.

        Notes
        -----
        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.
        """
        shape = stf.batch_shape + (len(sample_times),) + stf.value_shape
        return Tensor(create_operation(self.sample_stf, locals()), shape=shape)

    def identity_stf(self) -> Stf:
        r"""
        Create an Stf representing the identity function, f(t) = t.

        Returns
        -------
        Stf
            An Stf representing the identity function.

        See Also
        --------
        Graph.constant_stf: Create a batch of constant `Stf`\s.
        Graph.discretize_stf : Discretize an `Stf` into a `Pwc`.
        Graph.sample_stf : Sample an `Stf` at given times.

        Notes
        -----
        For more information on `Stf` nodes see the `Working with time-dependent functions in
        Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
        functions-in-boulder-opal>`_ topic.

        Examples
        --------
        Create Gaussian pulse.

        >>> time = graph.identity_stf()
        >>> time
        <Stf: operation_name="identity_stf", value_shape=(), batch_shape=()>
        >>> gaussian = graph.exp(- time ** 2)
        >>> gaussian
        <Stf: operation_name="exp", value_shape=(), batch_shape=()>

        See more examples in the `How to define continuous analytical Hamiltonians
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-define-continuous-analytical-
        hamiltonians>`_ user guide.
        """
        operation = create_operation(self.identity_stf, locals())
        return Stf(operation, value_shape=(), batch_shape=())



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from typing import (
    Any,
    Optional,
    Sequence,
    TypeVar,
)

import numpy as np

from boulderopal._nodes.node_data import Pwc
from boulderopal._validation import (
    ArrayT,
    Checker,
    ScalarT,
    nullable,
)
from boulderopal.graph import (
    Graph,
    execute_graph,
)
from boulderopal.ions._drives import (
    Drive,
    OptimizableDrive,
)
from boulderopal.optimization import run_optimization


def _validate_system_parameters(
    lamb_dicke_parameters: np.ndarray,
    relative_detunings: np.ndarray,
    target_phases: Optional[np.ndarray],
) -> int:
    """
    Validate the arrays describing an ion system
    and return the number of ions.
    """

    ion_count = lamb_dicke_parameters.shape[-1]

    Checker.VALUE(
        lamb_dicke_parameters.shape == (3, ion_count, ion_count)
        and relative_detunings.shape == (3, ion_count),
        "The shape of the Lamb–Dicke parameters array must be (3, N, N), "
        "and the shape of the relative detunings array must be (3, N), "
        "where N is the number of ions.",
        {
            "lamb_dicke_parameters.shape": lamb_dicke_parameters.shape,
            "relative_detunings.shape": relative_detunings.shape,
        },
    )

    if target_phases is not None:
        Checker.VALUE(
            target_phases.shape == (ion_count, ion_count),
            "The shape of the target phases array must be (N, N), "
            "where N is the number of ions.",
            {"target_phases.shape": target_phases.shape, "ion count": ion_count},
        )

    return ion_count


_T = TypeVar("_T", Drive, OptimizableDrive)


def _check_drives_addressing(drives: Sequence[_T], ion_count: int) -> None:
    """
    Check the input drives are a list and that the ions they address are valid.
    """

    Checker.TYPE(isinstance(drives, list), "You must provide a list of drives.")

    all_addressing: list[int] = []
    for idx, drive in enumerate(drives):
        Checker.VALUE(
            all(0 <= ion < ion_count for ion in drive.addressing),
            "The addressed ions must be between 0 (inclusive) "
            "and the number of ions (exclusive).",
            {f"drives[{idx}].addressing": drive.addressing, "ion count": ion_count},
        )
        all_addressing.extend(drive.addressing)

    Checker.VALUE(
        len(all_addressing) == len(set(all_addressing)),
        "Each ion can only be addressed by a single drive.",
    )


def _get_ion_drives(
    pwc_addressing_pairs: list[tuple[Pwc, tuple[int, ...]]], ion_count: int
) -> tuple[Sequence[Optional[Pwc]], list[int]]:
    """
    From a list of (Pwc, list(int)) tuples (drives and ions addressed by them),
    return (1) a list of length ion_count containing the drive addressing each ion
    or None if the ion is not addressed by any drive and (2) a list with the indices
    of the addressed ions.
    """
    ion_drives = [None] * ion_count
    addressed_ions = []
    for pwc, addressing in pwc_addressing_pairs:
        for ion in addressing:
            ion_drives[ion] = pwc  # type: ignore
            addressed_ions.append(ion)

    return ion_drives, addressed_ions


_MS_NODE_NAMES = ["sample_times", "phases", "displacements", "infidelities"]


def ms_simulate(
    drives: list[Drive],
    duration: float,
    lamb_dicke_parameters: np.ndarray,
    relative_detunings: np.ndarray,
    target_phases: Optional[np.ndarray] = None,
    sample_count: int = 128,
) -> dict:
    r"""
    Simulate a Mølmer–Sørensen-type operation on a system composed of ions.

    This function builds a graph describing the Mølmer–Sørensen operation
    and calls `boulderopal.execute_graph` to simulate the ion dynamics.

    Parameters
    ----------
    drives : list[~ions.Drive]
        A list of drives addressing the ions.
        Each ion can only be addressed by a single drive,
        but there may be ions not addressed by any drive.
    duration : float
        The duration, in seconds, of the dynamics to be simulated, :math:`T`.
        It must be greater than zero.
    lamb_dicke_parameters : np.ndarray
        The laser-ion coupling strength, :math:`\{\eta_{jkl}\}`.
        Its shape must be ``(3, N, N)``, where the dimensions indicate,
        respectively, axis, collective mode, and ion.
    relative_detunings : np.ndarray
        The difference :math:`\{\delta_{jk} = \nu_{jk} - \delta\}` (in Hz) between each motional
        mode frequency and the laser detuning from the qubit transition frequency :math:`\omega_0`.
        Its shape must be ``(3, N)``, where the dimensions indicate, respectively,
        axis and collective mode.
    target_phases : np.ndarray or None, optional
        The target total relative phases between ion pairs, :math:`\{\Psi_{ln}\}`,
        as a strictly lower triangular matrix of shape ``(N, N)``.
        :math:`\Psi_{ln}` with :math:`l > n` indicates the relative phase between ions
        :math:`l` and :math:`n`, while :math:`\Psi_{ln} = 0` for :math:`l \leq n`.
        If not provided, the function does not return the operational infidelities.
    sample_count : int, optional
        The number of times :math:`T` between 0 and `duration` (included)
        at which the evolution is sampled.
        Defaults to 128.

    Returns
    -------
    dict
        The result of the `execute_graph` call.
        Its ``output`` item is a dictionary containing information about
        the evolution of the system, with the following keys:

            ``sample_times``
                The times at which the evolution is sampled, as an array of shape ``(T,)``.
            ``phases``
                Acquired phases :math:`\{\Phi_{ln}(t_i) = \phi_{ln}(t_i) + \phi_{nl}(t_i)\}`
                for each sample time and for all ion pairs, as a strictly lower triangular
                matrix of shape ``(T, N, N)``.
                :math:`\Phi_{ln}(t_i)` with :math:`l > n` indicates the relative phase between ions
                :math:`l` and :math:`n`, while :math:`\Phi_{ln}(t_i) = 0` for :math:`l \leq n`.
            ``displacements``
                Displacements :math:`\{\eta_{jkl}\alpha_{jkl}(t_i)\}` for all mode-ion combinations,
                as an array of shape ``(T, 3, N, N)``, where the dimensions indicate, respectively,
                time, axis, collective mode, and ion.
            ``infidelities``
                A 1D array of length ``T`` with the operational infidelities of
                the Mølmer–Sørensen gate at each sample time, :math:`\mathcal{I}(t_i)`.
                Only returned if target relative phases are provided.

    See Also
    --------
    boulderopal.ions.Drive :
        Class describing non-optimizable drives.
    boulderopal.ions.ms_optimize :
        Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    boulderopal.ions.obtain_ion_chain_properties :
        Calculate the properties of an ion chain.

    Notes
    -----
    The internal and motional Hamiltonian of :math:`N` ions is

    .. math::
        H_0 = \sum_{j=1}^{3} \sum_{k=1}^{N} \hbar\nu_{jk} \left(a_{jk}^\dagger a_{jk}
            + \frac{1}{2}\right) + \sum_{l=1}^N \frac{\hbar \omega_0}{2} \sigma_{z,l} ,

    where :math:`j` indicates axis dimension (:math:`x`, :math:`y`, or :math:`z`),
    :math:`k` indicates collective mode, :math:`a_{jk}` is the annihilation operator,
    and :math:`\sigma_{z,l}` is the Pauli :math:`Z` operator for ion :math:`l`.

    The interaction Hamiltonian for Mølmer–Sørensen-type
    operations in the rotating frame with respect to :math:`H_0` is

    .. math::
        H_I(t) = i\hbar \sum_{j=1}^{3} \sum_{k=1}^{N} \sum_{l=1}^N
            \sigma_{x,l} \left(-\beta_{jkl}^*(t)a_{jk} + \beta_{jkl}(t) a_{jk}^\dagger\right) ,

    where :math:`\sigma_{x,l}` is the Pauli :math:`X` operator for ion :math:`l` and

    .. math::
        \beta_{jkl}(t) = \eta_{jkl} \frac{\gamma_l(t)}{2} \exp(i 2 \pi \delta_{jk} t)

    indicates the coupling between ion :math:`l` and motional mode :math:`(j,k)`.

    The corresponding unitary operation is given by [1]_

    .. math::
        U(t) = \exp\left[ \sum_{l=1}^N \sigma_{x,l} B_l(t)
            + i\sum_{l=1}^N\sum_{n=1}^{l-1} (\phi_{ln}(t) + \phi_{nl}(t))
            \sigma_{x,l} \sigma_{x,n} \right] ,

    where

    .. math::
        B_l(t) &\equiv \sum_{j=1}^{3} \sum_{k=1}^{N}
            \left(\eta_{jkl}\alpha_{jkl}(t)a_{jk}^\dagger
            - \eta_{jkl}^{\ast}\alpha_{jkl}^\ast(t)a_{jk} \right) ,

        \phi_{ln}(t) &\equiv \mathrm{Im} \left[ \sum_{j=1}^{3} \sum_{k=1}^{N}
            \int_{0}^{t} d \tau_1 \int_{0}^{\tau_1} d \tau_2
            \beta_{jkl}(\tau_1)\beta_{jkn}^{\ast}(\tau_2) \right] ,

        \alpha_{jkl}(t) &\equiv \int_0^t d\tau \frac{\gamma_l(\tau)}{2}
            \exp(i 2 \pi \delta_{jk} \tau) .

    The operational infidelity of the Mølmer–Sørensen gate is defined as [1]_:

    .. math::
        \mathcal{I} = 1 - \left| \left( \prod_{n=1}^N \prod_{l=n+1}^N
            \cos ( \Phi_{ln} - \Psi_{ln}) \right)
            \left( 1 - \sum_{j=1}^3 \sum_{k=1}^N \sum_{l=1}^N \left[ |\eta_{jkl} \alpha_{jkl}|^2
            \left(\bar{n}_{jk}+\frac{1}{2} \right) \right] \right) \right|^2 ,

    References
    ----------
    .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_
    """

    duration = ScalarT.REAL("duration").gt(0)(duration)

    lamb_dicke_parameters = ArrayT.REAL("lamb_dicke_parameters").ndim(3)(
        lamb_dicke_parameters
    )
    relative_detunings = ArrayT.REAL("relative_detunings").ndim(2)(relative_detunings)
    target_phases = nullable(ArrayT.REAL("target_phases").ndim(2), target_phases)
    ion_count = _validate_system_parameters(
        lamb_dicke_parameters, relative_detunings, target_phases
    )

    Checker.VALUE(
        all(isinstance(drive, Drive) for drive in drives),
        "All drives must be non-optimizable.",
        {"drives": drives},
    )

    _check_drives_addressing(drives, ion_count)

    graph = Graph()

    drive_pwcs = [
        (drive.get_pwc(graph, duration), drive.addressing) for drive in drives
    ]
    ion_drives, addressed_ions = _get_ion_drives(drive_pwcs, ion_count)

    sample_times = np.linspace(0.0, duration, sample_count)
    graph.tensor(sample_times, name="sample_times")

    phases = graph.ions.ms_phases(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[1],
    )

    displacements = graph.ions.ms_displacements(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
    )

    # If there are unaddressed ions add zeros for their displacements.
    ion_count = lamb_dicke_parameters.shape[-1]
    if len(addressed_ions) < ion_count:
        graph.matmul(
            displacements, np.eye(ion_count)[addressed_ions], name=_MS_NODE_NAMES[2]
        )
    else:
        displacements.name = _MS_NODE_NAMES[2]

    if target_phases is not None:
        graph.ions.ms_infidelity(
            phases=phases,
            displacements=displacements,
            target_phases=target_phases,
            name=_MS_NODE_NAMES[3],
        )
        output_node_names = _MS_NODE_NAMES
    else:
        output_node_names = _MS_NODE_NAMES[:3]

    return execute_graph(graph=graph, output_node_names=output_node_names)


def ms_optimize(
    drives: Sequence[OptimizableDrive],
    duration: float,
    lamb_dicke_parameters: np.ndarray,
    relative_detunings: np.ndarray,
    target_phases: np.ndarray,
    sample_count: int = 128,
    robust: bool = False,
    **optimization_kwargs: Any,
) -> dict:
    r"""
    Find optimal pulses to perform a target Mølmer–Sørensen-type operation
    on a system composed of ions.

    This function builds a graph describing the Mølmer–Sørensen operation
    and calls `boulderopal.run_optimization` to minimize the target cost.

    Parameters
    ----------
    drives : list[OptimizableDrive]
        A list of optimizable drives addressing the ions.
        Each ion can only be addressed by a single drive,
        but there may be ions not addressed by any drive.
    duration : float
        The duration, in seconds, of the dynamics to be optimized, :math:`T`.
        It must be greater than zero.
    lamb_dicke_parameters : np.ndarray
        The laser-ion coupling strength, :math:`\{\eta_{jkl}\}`.
        Its shape must be ``(3, N, N)``, where the dimensions indicate,
        respectively, axis, collective mode, and ion.
    relative_detunings : np.ndarray
        The difference :math:`\{\delta_{jk} = \nu_{jk} - \delta\}` (in Hz) between each motional
        mode frequency and the laser detuning from the qubit transition frequency :math:`\omega_0`.
        Its shape must be ``(3, N)``, where the dimensions indicate, respectively,
        axis and collective mode.
    target_phases : np.ndarray or None, optional
        The target total relative phases between ion pairs, :math:`\{\Psi_{ln}\}`,
        as a strictly lower triangular matrix of shape ``(N, N)``.
        :math:`\Psi_{ln}` with :math:`l > n` indicates the relative phase between ions
        :math:`l` and :math:`n`, while :math:`\Psi_{ln} = 0` for :math:`l \leq n`.
    sample_count : int, optional
        The number of times :math:`T` between 0 and `duration` (both included)
        at which the evolution is sampled.
        Defaults to 128.
    robust : bool, optional
        If set to False, the cost corresponds to the infidelity at the end of the gate.
        If set to True, the cost is the final infidelity plus a dephasing-robust cost term.
        Defaults to False.
    **optimization_kwargs
        Additional parameters to pass to boulderopal.run_optimization.

    Returns
    -------
    dict
        The result of the `run_optimization` call.
        Its ``output`` item is a dictionary containing information about
        the optimized drive and the evolution of the system, with the following keys:

            optimized drives
                The piecewise-constant optimized drives implementing the gate.
                The keys are the names of the `drives` provided to the function.
            ``sample_times``
                The times at which the evolution is sampled, as an array of shape ``(T,)``.
            ``phases``
                Acquired phases :math:`\{\Phi_{ln}(t_i) = \phi_{ln}(t_i) + \phi_{nl}(t_i)\}`
                for each sample time and for all ion pairs, as a strictly lower triangular
                matrix of shape ``(T, N, N)``.
                :math:`\Phi_{ln}(t_i)` with :math:`l > n` indicates the relative phase between ions
                :math:`l` and :math:`n`, while :math:`\Phi_{ln}(t_i) = 0` for :math:`l \leq n`.
            ``displacements``
                Displacements :math:`\{\eta_{jkl}\alpha_{jkl}(t_i)\}` for all mode-ion combinations,
                as an array of shape ``(T, 3, N, N)``, where the dimensions indicate, respectively,
                time, axis, collective mode, and ion.
            ``infidelities``
                A 1D array of length ``T`` with the operational infidelities of
                the Mølmer–Sørensen gate at each sample time, :math:`\mathcal{I}(t_i)`.

    See Also
    --------
    boulderopal.ions.ComplexOptimizableDrive :
        Class describing a piecewise-constant complex-valued optimizable drive.
    boulderopal.ions.RealOptimizableDrive :
        Class describing a piecewise-constant real-valued optimizable drive.
    boulderopal.ions.ms_simulate :
        Simulate a Mølmer–Sørensen-type operation on a trapped ions system.
    boulderopal.ions.obtain_ion_chain_properties :
        Calculate the properties of an ion chain.

    Notes
    -----
    See the notes of :func:`boulderopal.ions.ms_simulate` for the main equations and definitions.

    You can use the `robust` flag to construct a Mølmer–Sørensen gate that is
    robust against dephasing noise. This imposes a symmetry [1]_ in the optimizable
    ion drives and aims to minimize the time-averaged positions of the phase-space
    trajectories,

    .. math::
        \langle \alpha_{jkl} \rangle
            = \frac{1}{t_\text{gate}} \int_0^{t_\text{gate}}
                \alpha_{jkl}(t) \mathrm{d} t .

    This is achieved by adding an additional term to the cost function,
    consisting of the sum of the square moduli of the time-averaged positions
    multiplied by the corresponding Lamb–Dicke parameters. That is to say,

    .. math::
        C_\text{robust} = \mathcal{I} + \sum_{j=1}^{3} \sum_{k=1}^{N} \sum_{l=1}^{N}
                \left| \eta_{jkl} \langle \alpha_{jkl} \rangle \right|^2 .

    References
    ----------
    .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_
    """

    duration = ScalarT.REAL("duration").gt(0)(duration)

    lamb_dicke_parameters = ArrayT.REAL("lamb_dicke_parameters").ndim(3)(
        lamb_dicke_parameters
    )
    relative_detunings = ArrayT.REAL("relative_detunings").ndim(2)(relative_detunings)
    target_phases = ArrayT.REAL("target_phases").ndim(2)(target_phases)
    ion_count = _validate_system_parameters(
        lamb_dicke_parameters, relative_detunings, target_phases
    )

    Checker.TYPE(
        all(isinstance(drive, OptimizableDrive) for drive in drives),
        "All drives must be optimizable.",
    )

    _check_drives_addressing(drives, ion_count)

    drive_names = [drive.name for drive in drives]

    Checker.VALUE(
        len(drive_names) == len(set(drive_names)),
        "The drive names must be unique.",
        {"drive names": drive_names},
    )

    graph = Graph()

    drive_pwcs = [
        (drive.get_pwc(graph, duration, robust), drive.addressing) for drive in drives
    ]
    ion_drives, addressed_ions = _get_ion_drives(drive_pwcs, ion_count)

    sample_times = np.linspace(0.0, duration, sample_count)
    graph.tensor(sample_times, name="sample_times")

    phases = graph.ions.ms_phases(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[1],
    )

    displacements = graph.ions.ms_displacements(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
    )

    # If there are unaddressed ions add zeros for their displacements.
    ion_count = lamb_dicke_parameters.shape[-1]
    if len(addressed_ions) < ion_count:
        graph.matmul(
            displacements, np.eye(ion_count)[addressed_ions], name=_MS_NODE_NAMES[2]
        )
    else:
        displacements.name = _MS_NODE_NAMES[2]

    infidelities = graph.ions.ms_infidelity(
        phases=phases,
        displacements=displacements,
        target_phases=target_phases,
        name=_MS_NODE_NAMES[3],
    )

    cost = infidelities[-1]
    if robust:
        cost += graph.ions.ms_dephasing_robust_cost(
            drives=ion_drives,
            lamb_dicke_parameters=lamb_dicke_parameters,
            relative_detunings=relative_detunings,
        )

    return run_optimization(
        graph=graph,
        cost_node_name=cost.name,
        output_node_names=drive_names + _MS_NODE_NAMES,
        **optimization_kwargs,
    )



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Functionality related to the computational-flow graph object.
"""
from __future__ import annotations

from typing import Optional

import numpy as np
from qctrlcommons.graph import Graph as BaseGraph

from boulderopal._nodes.arithmetic_binary import ArithmeticBinaryGraph
from boulderopal._nodes.arithmetic_unary import ArithmeticUnaryGraph
from boulderopal._nodes.differentiation import DifferentiationGraph
from boulderopal._nodes.filter_function import FilterFunctionGraph
from boulderopal._nodes.fock_space import FockSpaceGraph
from boulderopal._nodes.infidelity import InfidelityGraph
from boulderopal._nodes.ions import IonsGraph
from boulderopal._nodes.node_data import (
    ConvolutionKernel,
    Pwc,
)
from boulderopal._nodes.optimization import OptimizationGraph
from boulderopal._nodes.oqs import OqsGraph
from boulderopal._nodes.pwc import PwcGraph
from boulderopal._nodes.random import RandomGraph
from boulderopal._nodes.signals import Signals
from boulderopal._nodes.sparse import SparseGraph
from boulderopal._nodes.stf import StfGraph
from boulderopal._nodes.tensor import TensorGraph
from boulderopal._validation import Checker


class Graph(  # pylint: disable=too-many-ancestors
    ArithmeticBinaryGraph,
    ArithmeticUnaryGraph,
    BaseGraph,
    DifferentiationGraph,
    FilterFunctionGraph,
    FockSpaceGraph,
    InfidelityGraph,
    OptimizationGraph,
    OqsGraph,
    PwcGraph,
    SparseGraph,
    StfGraph,
    TensorGraph,
):
    """
    A class for representing and building a Boulder Opal data flow graph.

    The graph object is the main entry point to the Boulder Opal graph ecosystem.
    You can call methods to add nodes to the graph, and use the `operations` attribute to get a
    dictionary representation of the graph.
    """

    def __init__(self) -> None:
        self.random = RandomGraph(self)
        self.signals = Signals(self)
        self.ions = IonsGraph(self)
        super().__init__()

    def filter_and_resample_pwc(
        self,
        pwc: Pwc,
        kernel: ConvolutionKernel,
        segment_count: int,
        duration: Optional[float] = None,
        *,
        name: Optional[str] = None,
    ) -> Pwc:
        r"""
        Filter a piecewise-constant function by convolving it with a kernel and resample it again.

        Parameters
        ----------
        pwc : Pwc
            The piecewise-constant function :math:`\alpha(t)` to be filtered.
        kernel : ConvolutionKernel
            The node representing the kernel :math:`K(t)`.
        segment_count : int
            The number of segments of the resampled filtered function.
        duration : float or None, optional
            Force the resulting piecewise-constant function to have a certain duration.
            This option is mainly to avoid floating point errors when the total duration is
            too small. Defaults to the sum of segment durations of `pwc`.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The filtered and resampled piecewise-constant function.

        See Also
        --------
        :func:`Graph.convolve_pwc` :
            Create the convolution of a piecewise-constant function with a kernel.
        :func:`Graph.discretize_stf` :
            Create a piecewise-constant function by discretizing a sampleable function.
        :func:`Graph.sinc_convolution_kernel` :
            Create a convolution kernel representing the sinc function.

        Notes
        -----
        The convolution is

        .. math::
            (\alpha * K)(t) \equiv
            \int_{-\infty}^\infty \alpha(\tau) K(t-\tau) \mathrm{d}\tau.

        Convolution in the time domain is equivalent to multiplication in the
        frequency domain, so this function can be viewed as applying a linear
        time-invariant filter (specified via its time domain kernel :math:`K(t)`)
        to :math:`\alpha(t)`.
        """
        return self.discretize_stf(
            stf=self.convolve_pwc(pwc=pwc, kernel=kernel),
            duration=duration or np.sum(pwc.durations),
            segment_count=segment_count,
            name=name,
        )

    def real_optimizable_pwc_signal(
        self,
        segment_count: int,
        duration: float,
        maximum: float,
        minimum: float = 0.0,
        initial_values: Optional[np.ndarray | list[np.ndarray]] = None,
        *,
        name: Optional[str] = None,
    ) -> Pwc:
        """
        Create a real optimizable piecewise-constant signal.

        Parameters
        ----------
        segment_count : int
            The number of piecewise-constant segments in the signal.
        duration : float
            The duration of the signal.
        maximum : float
            The upper bound for the signal values.
        minimum : float, optional
            The lower bound for the signal values. Defaults to 0.
        initial_values : np.ndarray or list[np.ndarray] or None, optional
            Initial values for the signal. You can either provide a single array,
            or a list of them. If a list of arrays is used, they must have the same length.
            Defaults to None, meaning the optimizer initializes the variables with random values.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The optimizable piecewise-constant signal.

        See Also
        --------
        :func:`Graph.complex_optimizable_pwc_signal` :
            Create a complex optimizable `Pwc` signal.
        :func:`Graph.optimization_variable` :
            Create optimization variables, which can be bounded, semi-bounded, or unbounded.
        :func:`Graph.pwc_signal` : Create a piecewise-constant signal.
        """
        Checker.VALUE(maximum > minimum, "maximum must be greater than minimum.")

        values = self.optimization_variable(
            count=segment_count,
            lower_bound=minimum,
            upper_bound=maximum,
            initial_values=initial_values,
        )
        return self.pwc_signal(values=values, duration=duration, name=name)

    def complex_optimizable_pwc_signal(
        self,
        segment_count: int,
        duration: float,
        maximum: float,
        initial_values: Optional[np.ndarray | list[np.ndarray]] = None,
        *,
        name: Optional[str] = None,
    ) -> Pwc:
        """
        Create a complex optimizable piecewise-constant signal.

        Parameters
        ----------
        segment_count : int
            The number of segments of the signal.
        duration : float
            The duration of the signal.
        maximum : float
            The upper bound for the modulus of the signal values.
        initial_values : np.ndarray or list[np.ndarray] or None, optional
            Initial values for the signal. You can either provide a single array,
            or a list of them. If a list of arrays is used, they must have the same length.
            Defaults to None, meaning the optimizer initializes the variables with random values.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The optimizable piecewise-constant signal.

        See Also
        --------
        :func:`Graph.real_optimizable_pwc_signal` :
            Create a real optimizable `Pwc` signal.
        :func:`Graph.complex_pwc_signal` :
            Create a complex piecewise-constant signal from moduli and phases.
        :func:`Graph.optimization_variable` :
            Create optimization variables, which can be bounded, semi-bounded, or unbounded.
        :func:`Graph.pwc_signal` : Create a piecewise-constant signal.

        Notes
        -----
        Note that this function sets limits to the modulus of the signal.

        If you want to set (different) limits to the real and imaginary parts instead,
        consider using `graph.real_optimizable_signal` to create signals for the
        real and imaginary parts, which you can pass to `graph.complex_value`.
        """

        Checker.VALUE(maximum > 0, "maximum must be greater than 0.")
        initial_moduli: Optional[list[np.ndarray] | np.ndarray] = None
        initial_phases: Optional[list[np.ndarray] | np.ndarray] = None

        if initial_values is not None:
            if isinstance(initial_values, list):
                initial_moduli = list(np.absolute(initial_values))
                initial_phases = list(np.angle(initial_values))
            else:
                initial_moduli = np.absolute(initial_values)
                initial_phases = np.angle(initial_values)

        moduli = self.optimization_variable(
            count=segment_count,
            lower_bound=0.0,
            upper_bound=maximum,
            initial_values=initial_moduli,
        )
        phases = self.optimization_variable(
            count=segment_count,
            lower_bound=-np.pi,
            upper_bound=np.pi,
            is_lower_unbounded=True,
            is_upper_unbounded=True,
            initial_values=initial_phases,
        )
        return self.complex_pwc_signal(
            moduli=moduli, phases=phases, duration=duration, name=name
        )



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from typing import (
    Any,
    List,
    Optional,
    Union,
)

import numpy as np

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.validation import (
    ShapeT,
    shapeable,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    ArrayT,
    Checker,
    ScalarT,
    pipe,
    sequence_like,
    validated,
)

from .node_data import Tensor


def _validate_fock_state_input(
    dimension: int | list[int],
    level: int | list[int] | list[list[int]],
    offset: Optional[int | list[int]],
) -> tuple[int, ...]:
    """
    Validate the inputs of the Fock state node and, if they're valid,
    return the shape of the Fock state.
    """
    if isinstance(dimension, list):
        # dimension is a list, so offset must be a list of non-negative integers of the same length.
        if offset is not None:
            Checker.VALUE(
                isinstance(offset, list) and len(offset) == len(dimension),
                "When constructed from subsystems, a non-default offset must be a list of integers "
                "and have the same size as dimension.",
                {"dimension": dimension, "offset": offset},
            )

        # level must be int or list[int] or list[list[int]].
        if isinstance(level, list):
            if isinstance(level[0], list):
                # level is list[list[int]].
                Checker.VALUE(
                    all(len(item) == len(dimension) for item in level),  # type: ignore
                    "Each batch of level must be a list of integers with the same "
                    "length as dimension.",
                    {"level": level, "dimension": dimension},
                )
            else:
                # level is list[int].
                Checker.VALUE(
                    len(level) == len(dimension),
                    "Level must be a list of integers with the same length as dimension.",
                    {"level": level, "dimension": dimension},
                )
            _level_arr = np.asarray(level)
        else:
            # level is int.
            _level_arr = np.repeat(level, len(dimension))

        # level must be in [offset, dim+offset).
        _dimension_arr = np.asarray(dimension)
        _offset_arr = (
            np.zeros_like(_dimension_arr) if offset is None else np.asarray(offset)
        )
        Checker.VALUE(
            np.all(_level_arr >= _offset_arr),
            "Level must be larger than or equal to offset.",
            {"level": level, "offset": offset},
        )
        Checker.VALUE(
            np.all(_level_arr < (_dimension_arr + _offset_arr)),
            "Level must be smaller than dimension + offset.",
            {"level": level, "dimension": dimension, "offset": offset},
        )

        return _level_arr.shape[:-1] + (int(np.prod(dimension)),)

    # dimension is an int, so offset must be a non-negative integer (if set).
    if offset is not None:
        if not isinstance(offset, int):
            raise TypeError(
                "Dimension and offset must be the same type. They should either be "
                "integers or lists of integers. "
                f"dimension={dimension}, offset={offset}."
            )
        _offset: int = offset
    else:
        _offset = 0

    # level must be an int or list[int] in [offset, dim+offset).
    Checker.VALUE(
        isinstance(level, int) or isinstance(level[0], int),
        "If the Fock state is not constructed from subsystems, "
        "level must either be an integer or a list of integers.",
        {"dimension": dimension, "level": level},
    )
    level_arr = np.asarray(level)
    Checker.VALUE(
        np.all(level_arr >= _offset),
        "Level must be larger than or equal to the offset.",
        {"offset": _offset, "level": level},
    )
    Checker.VALUE(
        np.all(level_arr < dimension + _offset),
        "Level must be smaller than dimension + offset.",
        {"dimension": dimension, "offset": _offset, "level": level},
    )

    return level_arr.shape + (dimension,)


def _validate_alpha(value: Any, *, name: str) -> Union[int, float, complex, np.ndarray]:
    try:
        return ScalarT.NUMERIC(name)(value)
    except TypeError:
        try:
            return ArrayT.NUMERIC().ndim(1)(value, name=name)
        except TypeError as exc:
            raise TypeError(f"{name} must be a number or a 1D array.") from exc


class FockSpaceGraph:  # pylint: disable=unused-argument
    """
    Base class implementing Fock space graph methods.
    """

    @validated
    def fock_state(
        self,
        dimension: Annotated[
            Union[int, List[int]], pipe(sequence_like(ScalarT.INT().ge(1)))
        ],
        level: Annotated[
            Union[int, List[int], List[List[int]]],
            pipe(sequence_like(sequence_like(ScalarT.INT().ge(0)))),
        ],
        offset: Optional[
            Annotated[Union[int, List[int]], pipe(sequence_like(ScalarT.INT().ge(0)))]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a Fock state (or a batch of them).

        You can create a Fock state for a single system by providing the `dimension` of the
        truncated Fock space, the occupied energy `level`, and the lowest represented level
        `offset` (optional) as integer parameters. You can create a batch of Fock states
        by passing a list of integers to `level`.

        Alternatively, you can create a Fock state for a composite system. In this case, the
        `dimension` (and the optional `offset`) must be a list of integers, representing
        the size of the truncated space for each subsystem (and the respective offset).
        You can then set the occupied level for each subsystem as follows:

        - `level` is an integer, meaning all subsystems have the same occupied level.
        - `level` is a list of integers, meaning each element represents the occupied level
          for the corresponding subsystem.
        - `level` is a list of lists. The outer list defines a batch of states.
          Each inner list element is treated as in the above case.

        Parameters
        ----------
        dimension : int or list[int]
            The size of the state representation in the truncated Fock space. A list of integers
            is interpreted as the sizes of subsystems, meaning the final Fock space has the same
            size as the product of these integers. By default, the Fock space is truncated as
            [0, dimension). If non-zero offset is passed, the space is then truncated at
            [offset, dimension + offset).
        level : int or list[int] or list[list[int]]
            The level at which the Fock basis is occupied. If a list of integers
            (or a list of integers when constructed from subsystems) is passed,
            this function returns a batch of Fock states, where each element is a Fock state
            with the energy level specified in the list.
        offset : int or list[int] or None, optional
            The lowest level of Fock state in the representation. Defaults to None, meaning
            the lowest level is 0. If set, this parameter must have the same type as `dimension`.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            Fock states with energy level specified by `level`. 1D vector if there is no batch
            in `level`, otherwise a 2D tensor where the first axis is the batch dimension.

        See Also
        --------
        Graph.annihilation_operator : Create an annihilation operator in the truncated Fock space.
        Graph.coherent_state : Create a coherent state (or a batch of them).
        Graph.creation_operator : Create a creation operator in the truncated Fock space.
        Graph.number_operator : Create a number operator in the truncated Fock space.

        Examples
        --------
        Create a Fock state for a single system.

        >>> graph.fock_state(2, 0, name="direct")
        <Tensor: name="direct", operation_name="fock_state", shape=(2,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="direct")
        >>> result["output"]["direct"]["value"]
        array([1.+0.j, 0.+0.j])

        Create a batch of Fock states for a single system.

        >>> graph.fock_state(2, [0, 1], name="direct_batch")
        <Tensor: name="direct_batch", operation_name="fock_state", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="direct_batch")
        >>> result["output"]["direct_batch"]["value"]
        array([[1.+0.j, 0.+0.j],
            [0.+0.j, 1.+0.j]])

        Create a batch of Fock states for a single system with an offset.

        >>> graph.fock_state(3, [1, 2], offset=1,  name="direct_offset")
        <Tensor: name="direct_offset", operation_name="fock_state", shape=(2, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="direct_offset")
        >>> result["output"]["direct_offset"]["value"]
        array([[1.+0.j, 0.+0.j, 0.+0.j],
            [0.+0.j, 1.+0.j, 0.+0.j]])

        Create a Fock state from subsystems.

        >>> graph.fock_state([2, 3], [1, 2], name="subsystems")
        <Tensor: name="subsystems", operation_name="fock_state", shape=(6,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="subsystems")
        >>> result["output"]["subsystems"]["value"]
        array([0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j])

        Create a batch of Fock states from subsystems.

        >>> graph.fock_state([2, 3], [[0, 1], [1, 2]], name="batch")
        <Tensor: name="batch", operation_name="fock_state", shape=(2, 6)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="batch")
        >>> result["output"]["batch"]["value"]
        array([[0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
            [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j]])

        Create a batch of Fock states with offset from subsystems.

        >>> graph.fock_state([2, 3], [[1, 4], [1, 3]], offset=[1, 2], name="offset")
        <Tensor: name="offset", operation_name="fock_state", shape=(2, 6)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="offset")
        >>> result["output"]["offset"]["value"]
        array([[0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
            [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j]])
        """
        shape = _validate_fock_state_input(dimension, level, offset)
        operation = create_operation(self.fock_state, locals())
        return Tensor(operation, shape=shape)

    @validated
    def creation_operator(
        self,
        dimension: Annotated[int, pipe(ScalarT.INT().ge(1))],
        offset: Annotated[int, pipe(ScalarT.INT().ge(0))] = 0,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a creation operator in the truncated Fock space.

        Parameters
        ----------
        dimension : int
            The size of the state representation in the truncated Fock space.
            By default, the Fock space is truncated at [0, dimension).
            If non-zero offset is passed, the space is then
            truncated at [offset, dimension + offset).
        offset : int, optional
            The lowest level of Fock state in the representation. Defaults to 0.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            A 2D tensor representing the creation operator.

        See Also
        --------
        Graph.annihilation_operator : Create an annihilation operator in the truncated Fock space.
        Graph.coherent_state : Create a coherent state (or a batch of them).
        Graph.fock_state : Create a Fock state (or a batch of them).
        Graph.number_operator : Create a number operator in the truncated Fock space.

        Examples
        --------
        Generate a creation operator for a two-level system.

        >>> graph.creation_operator(2, name="adagger")
        <Tensor: name="adagger", operation_name="creation_operator", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="adagger")
        >>> result["output"]["adagger"]["value"]
        array([[0.+0.j, 0.+0.j],
            [1.+0.j, 0.+0.j]])

        Apply a creation operator on the ground state such that
        :math:`a^\dagger|0\rangle = |1\rangle`.

        >>> adagger = graph.creation_operator(2)
        >>> state = adagger @ graph.fock_state(2, 0)[:, None]
        >>> state.name = "state"
        >>> result = bo.execute_graph(graph=graph, output_node_names="state")
        >>> result["output"]["state"]["value"]
        array([[0.+0.j],
            [1.+0.j]])

        Generate a creation operator for a three-level system with an offset.

        >>> graph.creation_operator(3, 1, name="adagger_offset")
        <Tensor: name="adagger_offset", operation_name="creation_operator", shape=(3, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="adagger_offset")
        >>> result["output"]["adagger_offset"]["value"]
        array([[0.+0.j, 0.+0.j, 0.+0.j],
            [1.41421356+0.j, 0.+0.j, 0.+0.j],
            [0.+0.j, 1.73205081+0.j, 0.+0.j]])

        Apply a creation operator with an offset such that
        :math:`a^\dagger|1\rangle = \sqrt{2}|2\rangle`.

        >>> adagger_offset = graph.creation_operator(3, 1)
        >>> state_offset = adagger_offset @ graph.fock_state(3, 1, 1)[:, None]
        >>> state.name = "offset"
        >>> result = bo.execute_graph(graph=graph, output_node_names="offset")
        >>> result["output"]["offset"]["value"]
        array([[0.        +0.j],
            [1.41421356+0.j],
            [0.        +0.j]])
        """
        operation = create_operation(self.creation_operator, locals())
        return Tensor(operation, shape=(dimension, dimension))

    @validated
    def annihilation_operator(
        self,
        dimension: Annotated[int, pipe(ScalarT.INT().ge(1))],
        offset: Annotated[int, pipe(ScalarT.INT().ge(0))] = 0,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create an annihilation operator in the truncated Fock space.

        Parameters
        ----------
        dimension : int
            The size of the state representation in the truncated Fock space.
            By default, the Fock space is truncated at [0, dimension).
            If non-zero offset is passed, the space is then
            truncated at [offset, dimension + offset).
        offset : int, optional
            The lowest level of Fock state in the representation. Defaults to 0.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            A 2D tensor representing the annihilation operator.

        See Also
        --------
        Graph.coherent_state : Create a coherent state (or a batch of them).
        Graph.creation_operator : Create a creation operator in the truncated Fock space.
        Graph.fock_state : Create a Fock state (or a batch of them).
        Graph.number_operator : Create a number operator in the truncated Fock space.

        Examples
        --------
        Generate an annihilation operator for a two-level system.

        >>> graph.annihilation_operator(2, name="a")
        <Tensor: name="a", operation_name="annihilation_operator", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="a")
        >>> result["output"]["a"]["value"]
        array([[0.+0.j, 1.+0.j],
            [0.+0.j, 0.+0.j]])

        Apply an annihilation operator on the excited state such that
        :math:`a|1\rangle = |0\rangle`.

        >>> a = graph.annihilation_operator(2)
        >>> state = a @ graph.fock_state(2, 1)[:, None]
        >>> state.name = "state"
        >>> result = bo.execute_graph(graph=graph, output_node_names="state")
        >>> result["output"]["state"]["value"]
        array([[1.+0.j],
            [0.+0.j]])

        Generate an annihilation operator for a three-level system with an offset.

        >>> graph.annihilation_operator(3, 1, name="a_offset")
        <Tensor: name="a_offset", operation_name="annihilation_operator", shape=(3, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="a_offset")
        >>> result["output"]["a_offset"]["value"]
        array([[0.+0.j, 1.41421356+0.j, 0.+0.j],
            [0.+0.j, 0.+0.j, 1.73205081+0.j],
            [0.+0.j, 0.+0.j, 0.+0.j]])

        Apply an annihilation operator with an offset such that
        :math:`a|2\rangle = \sqrt{2}|1\rangle`.

        >>> a_offset = graph.creation_operator(3, 1)
        >>> state_offset = a_offset @ graph.fock_state(3, 2, 1)[:, None]
        >>> state.name = "offset"
        >>> result = bo.execute_graph(graph=graph, output_node_names="offset")
        >>> result["output"]["offset"]["value"]
        array([[1.41421356+0.j],
            [0.        +0.j],
            [0.        +0.j]])
        """
        operation = create_operation(self.annihilation_operator, locals())
        return Tensor(operation, shape=(dimension, dimension))

    @validated
    def number_operator(
        self,
        dimension: Annotated[int, pipe(ScalarT.INT().ge(1))],
        offset: Annotated[int, pipe(ScalarT.INT().ge(0))] = 0,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a number operator in the truncated Fock space.

        Parameters
        ----------
        dimension : int
            The size of the state representation in the truncated Fock space.
            By default, the Fock space is truncated at [0, dimension).
            If non-zero offset is passed, the space is then
            truncated at [offset, dimension + offset).
        offset : int, optional
            The lowest level of Fock state in the representation. Defaults to 0.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            A 2D tensor representing the number operator.

        See Also
        --------
        Graph.annihilation_operator : Create an annihilation operator in the truncated Fock space.
        Graph.coherent_state : Create a coherent state (or a batch of them).
        Graph.creation_operator : Create a creation operator in the truncated Fock space.
        Graph.fock_state : Create a Fock state (or a batch of them).

        Examples
        --------
        Generate a number operator for a three-level system.

        >>> graph.number_operator(3, name="n")
        <Tensor: name="n", operation_name="number_operator", shape=(3, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="n")
        >>> result["output"]["n"]["value"]
        array([[0.+0.j, 0.+0.j, 0.+0.j],
            [0.+0.j, 1.+0.j, 0.+0.j],
            [0.+0.j, 0.+0.j, 2.+0.j]])

        Apply a number operator on the second excited state such that
        :math:`N|2\rangle = 2|2\rangle`.

        >>> n = graph.number_operator(3)
        >>> state = n @ graph.fock_state(3, 2)[:, None]
        >>> state.name = "state"
        >>> result = bo.execute_graph(graph=graph, output_node_names="state")
        >>> result["output"]["state"]["value"]
        array([[0.+0.j],
            [0.+0.j],
            [2.+0.j]])

        Generate a number operator for a three-level system with an offset.

        >>> graph.number_operator(3, 1, name="n_offset")
        <Tensor: name="n_offset", operation_name="number_operator", shape=(3, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="n_offset")
        >>> result["output"]["n_offset"]["value"]
        array([[1.+0.j, 0.+0.j, 0.+0.j],
            [0.+0.j, 2.+0.j, 0.+0.j],
            [0.+0.j, 0.+0.j, 3.+0.j]])

        Apply a number operator with an offset such that
        :math:`N|3\rangle = 3|3\rangle`.

        >>> n_offset = graph.number_operator(3, 1)
        >>> state_offset = n_offset @ graph.fock_state(3, 3, 1)[:, None]
        >>> state_offset.name = "state_offset"
        >>> result = bo.execute_graph(graph=graph, output_node_names="state_offset")
        >>> result["output"]["state_offset"]["value"]
        array([[0.+0.j],
            [0.+0.j],
            [3.+0.j]])
        """
        operation = create_operation(self.number_operator, locals())
        return Tensor(operation, shape=(dimension, dimension))

    @validated
    def coherent_state(
        self,
        alpha: Annotated[Union[int, float, complex, np.ndarray], pipe(_validate_alpha)],
        dimension: Annotated[int, pipe(ScalarT.INT().ge(1))],
        offset: Annotated[int, pipe(ScalarT.INT().ge(0))] = 0,
        from_displacement: bool = True,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a coherent state (or a batch of them).

        By default, this function generates the coherent state by applying the displacement operator
        on the vacuum state in the truncated Fock space. Alternatively, you can also generate the
        coherent state from its analytical formula in the Fock basis (see the note part for details)
        by setting the flag `from_displacement` to False. Note that as the second approach is
        effectively truncating an infinite series, therefore the returned coherent state is not
        necessarily normalized.

        Parameters
        ----------
        alpha : number or np.ndarray
            A number :math:`\alpha` that characterizes the coherent state.
            You can also pass a 1D array to generate a batch of coherent states.
        dimension : int
            The size of the state representation in the truncated Fock space.
            By default, the Fock space is truncated at [0, dimension).
            If non-zero offset is passed, the space is then
            truncated at [offset, dimension + offset).
        offset : int, optional
            The lowest level of Fock state in the representation. Defaults to 0.
        from_displacement :  bool, optional
            Defaults to True, meaning the coherent state is computed from the displacement
            operation. You can set this flag to False to get the coherent state from its
            analytical formula in the Fock basis.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            A tensor representing the coherent state. The shape is 1D if there is
            no batch in `alpha`, otherwise 2D where the first axis is the batch dimension.

        See Also
        --------
        Graph.annihilation_operator : Create an annihilation operator in the truncated Fock space.
        Graph.creation_operator : Create a creation operator in the truncated Fock space.
        Graph.displacement_operator :
            Create a displacement operator (or a batch of them) in the truncated Fock space.
        Graph.fock_state : Create a Fock state (or a batch of them).
        Graph.number_operator : Create a number operator in the truncated Fock space.

        Notes
        -----
        A coherent state :math:`|\alpha\rangle` can be generated by applying a displacement
        operator :math:`D(\alpha)` on the vacuum state :math:`|0\rangle`:

        .. math::
            |\alpha\rangle = D(\alpha)|0\rangle
                        = \exp(\alpha \hat{a}^\dagger  - \alpha^\ast \hat{a}) | 0 \rangle

        where :math:`\hat{a}` and :math:`\hat{a}^\dagger` are the annihilation
        and creation operators respectively.

        The coherent state can also be represented in the basis of Fock states:

        .. math::
            |\alpha\rangle = e^{-\frac{|\alpha|^2}{2}}
                            \sum_{n=0}^{\infty} \frac{\alpha^n}{\sqrt{n!}}|n\rangle .

        For more information about coherent states, see `coherent state`_ on Wikipedia.

        .. _coherent state:
            https://en.wikipedia.org/wiki/Coherent_state

        Examples
        --------
        Create a coherent state from displacement.

        >>> graph.coherent_state(1j, 2, name="state")
        <Tensor: name="state", operation_name="coherent_state", shape=(2,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="state")
        >>> result["output"]["state"]["value"]
        array([5.40302306e-01+0.j        , 5.55111512e-17+0.84147098j])

        Create a batch of coherent states from displacement.

        >>> graph.coherent_state([1j, 3], 2, name="batch")
        <Tensor: name="batch", operation_name="coherent_state", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="batch")
        >>> result["output"]["batch"]["value"]
        array([[ 5.40302306e-01+0.j        ,  5.55111512e-17+0.84147098j],
            [-9.89992497e-01+0.j        ,  1.41120008e-01+0.j        ]])

        Create a batch of coherent states from the analytical formula.

        >>> graph.coherent_state([1j, 3], 2, from_displacement=False, name="analytical")
        <Tensor: name="analytical", operation_name="coherent_state", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="analytical")
        >>> result["output"]["analytical"]["value"]
        array([[6.06530660e-01+0.j        , 3.71392916e-17+0.60653066j],
            [1.11089965e-02+0.j        , 3.33269896e-02+0.j        ]])
        """
        batch_shape = getattr(alpha, "shape", ())
        operation = create_operation(self.coherent_state, locals())
        return Tensor(operation, shape=batch_shape + (dimension,))

    @validated
    def displacement_operator(
        self,
        alpha: Annotated[Union[int, float, complex, np.ndarray], pipe(_validate_alpha)],
        dimension: Annotated[int, pipe(ScalarT.INT().ge(1))],
        offset: Annotated[int, pipe(ScalarT.INT().ge(0))] = 0,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a displacement operator (or a batch of them).

        Parameters
        ----------
        alpha : number or np.ndarray
            A number :math:`\alpha` that characterizes the displacement operator.
            You can also pass a 1D array to generate a batch of displacement operators.
        dimension : int
            The size of the truncated Fock space.
            By default, the Fock space is truncated at [0, dimension).
            If non-zero offset is passed, the space is then
            truncated at [offset, dimension + offset).
        offset : int, optional
            The lowest Fock state.
            Defaults to 0.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            A tensor representing the displacement operator. The shape is 2D if there is no batch in
            `alpha`, otherwise 3D where the first axis is the batch dimension.

        See Also
        --------
        Graph.annihilation_operator : Create an annihilation operator in the truncated Fock space.
        Graph.coherent_state : Create a coherent state (or a batch of them).
        Graph.creation_operator : Create a creation operator in the truncated Fock space.
        Graph.fock_state : Create a Fock state (or a batch of them).
        Graph.number_operator : Create a number operator in the truncated Fock space.
        Graph.squeeze_operator :
            Create a squeeze operator (or a batch of them) in the truncated Fock space.

        Notes
        -----
        The displacement operator is defined as:

        .. math::
            D(\alpha) = \exp(\alpha \hat{a}^\dagger  - \alpha^\ast \hat{a})

        where :math:`\hat{a}` and :math:`\hat{a}^\dagger` are the annihilation
        and creation operators respectively.

        This operation uses truncated annihilation/creation operators to calculate the displacement
        operator. As such, for nonzero offset, generating a coherent state by applying the
        displacement operator to a Fock state will not generate the same result as using
        `coherent_state`, which uses a full operator and then truncates the resulting state.
        As the energy levels increase, however, the two coherent states will increasingly agree.

        For more information about the displacement operator, see `Displacement operator`_
        on Wikipedia.

        .. _Displacement operator:
            https://en.wikipedia.org/wiki/Displacement_operator

        Examples
        --------
        Create a displacement operator.

        >>> graph.displacement_operator(1j, 2, name="displacement")
        <Tensor: name="displacement", operation_name="displacement_operator", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="displacement")
        >>> result["output"]output["displacement"]["value"]
        array([[ 5.40302306e-01+0.00000000e+00j, -1.03050475e-16+8.41470985e-01j],
            [ 5.55111512e-17+8.41470985e-01j,  5.40302306e-01+6.61679489e-17j]])

        Create a batch of displacement operators.

        >>> graph.displacement_operator([1j, 3], 2, name="displacement_batch")
        <Tensor: name="displacement_batch", operation_name="displacement_operator", shape=(2, 2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="displacement_batch")
        >>> result["output"]["displacement_batch"]["value"]
        array([[[ 5.40302306e-01+0.00000000e+00j, -1.11022302e-16+8.41470985e-01j],
                [ 5.55111512e-17+8.41470985e-01j, 5.40302306e-01+1.11022302e-16j]],
            [[-9.89992497e-01+0.00000000e+00j, -1.41120008e-01+0.00000000e+00j],
                [ 1.41120008e-01+0.00000000e+00j, -9.89992497e-01+0.00000000e+00j]]])
        """
        batch_shape = getattr(alpha, "shape", ())
        operation = create_operation(self.displacement_operator, locals())
        return Tensor(operation, shape=batch_shape + (dimension, dimension))

    @validated
    def squeeze_operator(
        self,
        zeta: Annotated[Union[int, float, complex, np.ndarray], pipe(_validate_alpha)],
        dimension: Annotated[int, pipe(ScalarT.INT().ge(1))],
        offset: Annotated[int, pipe(ScalarT.INT().ge(0))] = 0,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a squeeze operator (or a batch of them).

        Parameters
        ----------
        zeta : number or np.ndarray
            A number :math:`\zeta` that characterizes the squeeze operator.
            You can also pass a 1D array to generate a batch of squeeze operators.
        dimension : int
            The size of the truncated Fock space.
            By default, the Fock space is truncated at [0, dimension).
            If non-zero offset is passed, the space is then
            truncated at [offset, dimension + offset).
        offset : int, optional
            The lowest Fock state.
            Defaults to 0.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            A tensor representing the squeeze operator. The shape is 2D if there is no batch in
            `zeta`, otherwise 3D where the first axis is the batch dimension.

        See Also
        --------
        Graph.annihilation_operator : Create an annihilation operator in the truncated Fock space.
        Graph.coherent_state : Create a coherent state (or a batch of them).
        Graph.creation_operator : Create a creation operator in the truncated Fock space.
        Graph.displacement_operator :
            Create a displacement operator (or a batch of them) in the truncated Fock space.
        Graph.fock_state : Create a Fock state (or a batch of them).
        Graph.number_operator : Create a number operator in the truncated Fock space.

        Notes
        -----
        The squeeze operator is defined as:

        .. math::
            S(\zeta) = \exp\left( \frac{\zeta^\ast \hat{a}^2 - \zeta \hat{a}^{\dagger2}}{2} \right)

        where :math:`\hat{a}` and :math:`\hat{a}^\dagger` are the annihilation
        and creation operators respectively.

        For more information about the squeeze operator, see `Squeeze operator`_ on Wikipedia.

        .. _Squeeze operator:
            https://en.wikipedia.org/wiki/Squeeze_operator

        Examples
        --------
        Create a squeeze operator.

        >>> graph.squeeze_operator(1j, 3, name="squeeze")
        <Tensor: name="squeeze", operation_name="squeeze_operator", shape=(3, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="squeeze")
        >>> result["output"]["squeeze"]["value"]
        array([[0.7602446+0.j        , 0.       +0.j        ,0.       -0.64963694j],
            [0.       +0.j        , 1.       +0.j        ,0.       +0.j        ],
            [0.       -0.64963694j, 0.       +0.j        ,0.7602446+0.j        ]])

        Create a batch of squeeze operators.

        >>> graph.squeeze_operator([1j, 3], 3, name="squeeze_batch")
        <Tensor: name="squeeze_batch", operation_name="squeeze_operator", shape=(2, 3, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="squeeze_batch")
        >>> result["output"]["squeeze_batch"]["value"]
        array([[[ 0.7602446 +0.j        ,  0.        +0.j        ,0.        -0.64963694j],
                [ 0.        +0.j        ,  1.        +0.j        ,0.        +0.j        ],
                [ 0.        -0.64963694j,  0.        +0.j        ,0.7602446 +0.j        ]],
            [[-0.19569944+0.j        ,  0.        +0.j        ,0.98066392+0.j        ],
                [ 0.        +0.j        ,  1.        +0.j        ,0.        +0.j        ],
                [-0.98066392-0.j        ,  0.        -0.j        ,-0.19569944-0.j        ]]])
        """
        batch_shape = getattr(zeta, "shape", ())
        operation = create_operation(self.squeeze_operator, locals())
        return Tensor(operation, shape=batch_shape + (dimension, dimension))

    @validated
    def wigner_transform(
        self,
        density_matrix: Annotated[
            Union[Tensor, np.ndarray], pipe(shapeable, after=ShapeT.OPERATOR())
        ],
        position: Annotated[np.ndarray, pipe(ArrayT.NUMERIC().ndim(1))],
        momentum: Annotated[np.ndarray, pipe(ArrayT.NUMERIC().ndim(1))],
        offset: Annotated[int, pipe(ScalarT.INT().ge(0))] = 0,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Transform a density matrix into a Wigner function (or a batch of them).

        Parameters
        ----------
        density_matrix : Tensor or np.ndarray
            The density matrix :math:`\rho` in the Fock basis.
            Must be of shape ``(..., D, D)``.
        position : np.ndarray
            The dimensionless position vector :math:`q`. Must be a 1D array of shape ``(L,)``.
        momentum : np.ndarray
            The dimensionless momentum vector :math:`p`. Must be a 1D array of shape ``(K,)``.
        offset : int, optional
            The lowest Fock state. Defaults to 0.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The Wigner function with shape ``(..., L, K)``.

        Warnings
        --------
        This function currently does not support calculating a gradient with respect to its inputs.
        Therefore, it cannot be used in a graph for a `run_optimization` or
        `run_stochastic_optimization` call, which will raise a `RuntimeError`.
        Please use gradient-free optimization if you want to perform an optimization task with this
        function. You can learn more about it in the
        `How to optimize controls using gradient-free optimization
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-using-gradient-free-optimization>`_
        user guide.

        See Also
        --------
        Graph.coherent_state : Create a coherent state (or a batch of them).
        Graph.fock_state : Create a Fock state (or a batch of them).

        Notes
        -----
        The Wigner transform is defined as:

        .. math::
            W(q,p) = \frac{1}{2\pi} \int_{-\infty}^{\infty} \mathrm{e}^{\mathrm{i}sp}
            \langle q - s/2| \rho | q + s/2 \rangle \mathrm{d}s .

        For more information about the Wigner transform, see `Wigner function`_ on Wikipedia.

        .. _Wigner function:
            https://en.wikipedia.org/wiki/Wigner_quasiprobability_distribution

        Examples
        --------
        Create a Wigner function.

        >>> graph.wigner_transform(np.array([[1]]), [-1, 0, 1], np.zeros(1), name="wigner")
        <Tensor: name="wigner", operation_name="wigner_transform", shape=(3, 1)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="wigner")
        >>> result["output"]["wigner"]["value"]
        array([[ 0.11709966+0.j],
            [ 0.31830989+0.j],
            [ 0.11709966+0.j]])

        Create a batch of Wigner function.

        >>> graph.wigner_transform(
                np.array([[[1, 0], [0, 0]], [[0, 0], [0, 1]]]),
                np.array([-1, 0, 1]),
                np.zeros(1),
                name="wigner_batch",
            )
        <Tensor: name="wigner_batch", operation_name="wigner_transform", shape=(2, 3, 1)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="wigner_batch")
        >>> result["output"]["wigner_batch"]["value"]
        array([[[ 0.11709966+0.j],
                [ 0.31830989+0.j],
                [ 0.11709966+0.j]],
            [[ 0.11709966+0.j],
                [-0.31830989+0.j],
                [ 0.11709966+0.j]]])
        """
        batch_shape = density_matrix.shape[:-2]
        operation = create_operation(self.wigner_transform, locals())
        return Tensor(operation, shape=batch_shape + (len(position), len(momentum)))



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import (
    Callable,
    Optional,
)

import numpy as np

from boulderopal._nodes.node_data import (
    Pwc,
    SparsePwc,
    Tensor,
)
from boulderopal._validation import Checker


def get_broadcasted_shape(*shapes: tuple[int, ...], message: str) -> tuple[int, ...]:
    """
    Return the shape resulting of broadcasting multiple shapes,
    or None if they're not broadcastable.

    The shapes are broadcastable if, for each dimension starting from the end,
    they all have either the same size or a size 1.

    Parameters
    ----------
    *shapes : tuple[int]
        Shapes of the objects.
    message : str
        The error message.

    Returns
    -------
    tuple[int]
        The resulting broadcasted shape.
    """
    try:
        return np.broadcast_shapes(*shapes)
    except ValueError as e:
        raise ValueError(message) from e


def validate_function_output_shapes(
    x_batch_shape: tuple[int, ...],
    x_value_shape: tuple[int, ...],
    y_batch_shape: tuple[int, ...],
    y_value_shape: tuple[int, ...],
    validate_value_shape: Callable,
) -> tuple[tuple[int, ...], tuple[int, ...]]:
    """
    Get the output batch and value shape for two input shapes of Pwcs/Stfs.
    The names of the variables are assumed to be x and y when reporting errors.

    Parameters
    ----------
    x_batch_shape : tuple[int]
        The batch shape of the first object.
    x_value_shape : tuple[int]
        The value shape of the first object.
    y_batch_shape : tuple[int]
        The batch shape of the second object.
    y_value_shape : tuple[int]
        The value shape of the second object.
    validate_value_shape : Callable[[tuple, tuple, str, str], tuple]
        Function that takes the value shapes of two Tensors, Pwcs,
        or Stfs (as well as their names), and returns the expected values
        shape of the output Tensor, Pwc, or Stf. The function
        shouldn't assume that the shapes are compatible, and raise an
        exception if they aren't. The names provided should be used to
        generate the error message.

    Returns
    -------
    tuple[int], tuple[int]
        The batch and value shapes of the output Pwc/Stf.
    """
    batch_shape = get_broadcasted_shape(
        x_batch_shape,
        y_batch_shape,
        message="The batch shapes of x and y must be broadcastable.",
    )
    value_shape = validate_value_shape(x_value_shape, y_value_shape, "x", "y")

    return batch_shape, value_shape


def validate_tensor_and_function_output_shapes(
    t_shape: tuple[int, ...],
    f_batch_shape: tuple[int, ...],
    f_value_shape: tuple[int, ...],
    t_name: str,
    f_name: str,
    validate_value_shape: Callable,
    tensor_first: bool = True,
) -> tuple[tuple[int, ...], tuple[int, ...]]:
    """
    Get the output batch and value shape for an input tensor and an input Pwc/Stf.

    Parameters
    ----------
    t_shape : tuple[int, ...]
        The shape of the tensor.
    f_batch_shape : tuple[int, ...]
        The batch shape of the Pwc/Stf.
    f_value_shape : tuple[int, ...]
        The value shape of the Pwc/Stf.
    t_name : str
        The name of the tensor variable, used for the error message in case the shapes aren't
        compatible.
    f_name : str
        The name of the function variable, used for the error message in case the shapes aren't
        compatible.
    validate_value_shape : Callable[[tuple, tuple, str, str], tuple]
        Function that takes the value shapes of two Tensors, Pwcs,
        or Stfs (as well as their names), and returns the expected values
        shape of the output Tensor, Pwc, or Stf. The function
        shouldn't assume that the shapes are compatible, and raise an
        exception if they aren't. The names provided should be used to
        generate the error message.
    tensor_first : bool, optional
        Whether the Tensor is the leftmost parameter. Defaults to True.

    Returns
    -------
    tuple[int, ...], tuple[int, ...]
        The batch and value shapes of the output Pwc/Stf.
    """
    if tensor_first:
        value_shape = validate_value_shape(t_shape, f_value_shape, t_name, f_name)
    else:
        value_shape = validate_value_shape(f_value_shape, t_shape, f_name, t_name)

    return f_batch_shape, value_shape


def check_operation_axis(
    axis: Optional[list[int] | int], shape: tuple[int, ...], tensor_name: str
) -> list[int]:
    """
    Certain Tensor operations are applied along the axis of the Tensor.
    The function checks
    1. whether the axis is consistent with the shape of the tensor.
    2. whether there are any repeated items in axis.
    """

    if axis is None:
        return list(range(len(shape)))

    if not isinstance(axis, list):
        axis = [axis]

    for i, dimension in enumerate(axis):
        Checker.VALUE(
            -len(shape) <= dimension < len(shape),
            f"Elements of axis must be valid axes of {tensor_name} (between {-len(shape)} "
            f"and {len(shape)-1}, inclusive).",
            {f"axis[{i}]": dimension, f"len({tensor_name}.shape)": len(shape)},
        )
        if dimension < 0:
            axis[i] = dimension + len(shape)

    Checker.VALUE(
        len(set(axis)) == len(axis), "Elements of axis must be unique.", {"axis": axis}
    )

    return axis


def get_keepdims_operation_shape(
    shape: tuple[int, ...], axis: list[int], keepdims: bool
) -> tuple[int, ...]:
    """
    Return the shape of the operations that can keep the dimension of the input tensor.
    """
    output_shape = []
    for i, size in enumerate(shape):
        if i not in axis:
            output_shape.append(size)
        elif keepdims:
            output_shape.append(1)
    return tuple(output_shape)


def mesh_pwc_durations(pwcs: list[Pwc] | list[SparsePwc]) -> np.ndarray:
    """
    Return an array with the durations resulting of meshing the durations
    of the input PWC functions.

    Parameters
    ----------
    pwcs : list[Pwc] or list[SparsePwc]
        The Pwc functions whose durations should be meshed.

    Returns
    -------
    np.array
        The array with meshed durations.
    """
    _durations = [sum(pwc.durations) for pwc in pwcs]
    Checker.VALUE(
        np.allclose(_durations[0], _durations, atol=0.0),
        "All Pwc must have the same duration.",
    )

    times = np.unique(np.concatenate([np.cumsum(pwc.durations) for pwc in pwcs]))
    return np.diff(np.insert(times, 0, 0))


def get_arraylike_shape(
    x: Tensor | float | int | complex | np.ndarray,
) -> tuple[int, ...]:
    """
    Return the shape of an array-like object.
    """
    return getattr(x, "shape", ())
